from unittest import mock
import typing

from karp5.dbhandler.dbhandler import DBUsageError, SQLNull, get_entries_to_keep_gen


def test_sql_null_w_lexicon():
    sql_null = SQLNull("LEXICON")
    assert str(sql_null) == "Error: No SQL db available for LEXICON"
    assert sql_null.message == "Error: No SQL db available for LEXICON"


def test_db_usage_error():
    msg = "TEST"
    db_usage_error = DBUsageError(msg)

    assert db_usage_error.message == f"Error: {msg}"


def generate_from_list(
    l: typing.List[typing.Dict[str, typing.Any]]
) -> typing.Generator[typing.Dict[str, typing.Any], None, None]:
    for obj in l:
        yield obj


def test_get_entries_to_keep_gen():
    lexicon = "foo"
    entries = [
        {"id": 1, "value": "expected", "status": "updated"},
        {"id": 1, "value": "not expected", "status": "added"},
        {"id": 2, "value": "not expected", "status": "removed"},
        {"id": 3, "value": "expected", "status": "added"},
        {"id": 3, "value": "not expected", "status": "removed"},
    ]

    with mock.patch(
        "karp5.dbhandler.dbhandler.dbselect_gen", return_value=generate_from_list(entries)
    ), mock.patch("karp5.dbhandler.dbhandler.get_engine", return_value=("ENGINE", "DBENTRY")):
        for entry in get_entries_to_keep_gen(lexicon):
            assert entry["value"] == "expected"



# -*- coding: utf-8 -*-
""" Methods for querying the data base """
import copy
import itertools
from itertools import chain
import json
import logging
import re

import elasticsearch
from flask import Response, jsonify, request, stream_with_context

from karp5 import errors
from karp5 import util
from karp5.config import mgr as conf_mgr
from karp5.context import auth
import karp5.dbhandler.dbhandler as db
import karp5.server.helper.helpers as helpers
from karp5.server.translator import parser
from karp5.server.translator.errors import AuthenticationError, QueryError

_logger = logging.getLogger("karp5")


def query(page=0):
    try:
        ans = requestquery(page=page)
        return jsonify(ans)

    except errors.KarpException as e:  # pass on karp exceptions
        _logger.exception(e)
        raise
    except Exception as e:  # catch *all* exceptions and show for user
        _logger.exception(e)
        raise errors.KarpGeneralError(str(e), user_msg=str(e), query=request.query_string)


def requestquery(page=0):
    """ The Function for querying our database """
    _logger.debug("|requestquery| page = %d", page)
    # page is assumed to be 0 indexed here
    user_is_authorized, permitted = auth.validate_user(mode="read")
    try:
        # default values
        default = {"size": 25, "page": page, "version": "true"}
        settings = parser.make_settings(permitted, default, user_is_authorized=user_is_authorized)
        elasticq = parser.parse(settings)
    except QueryError as e:
        _logger.exception(e)
        raise errors.KarpQueryError(
            "Parse error - " + e.message, debug_msg=e.debug_msg, query=request.query_string,
        )
    except AuthenticationError as e:
        _logger.exception(e)
        msg = e.message
        raise errors.KarpAuthenticationError(msg)
    except errors.KarpException as e:  # pass on karp exceptions
        _logger.exception(e)
        raise
    except Exception as e:  # catch *all* exceptions
        _logger.exception(e)
        raise errors.KarpQueryError("Could not parse data", debug_msg=e, query=request.query_string)
    mode = settings["mode"]
    sort = sortorder(settings, mode, settings.get("query_command", ""))
    if "start" in settings:
        start = settings["start"]
    else:
        start = settings["page"] * settings["size"]

    # size = min(settings['size'], setupconf.max_page)
    size = settings["size"]
    index, typ = conf_mgr.get_mode_index(mode)
    exclude = conf_mgr.searchfield(mode, "secret_fields") if not user_is_authorized else []
    ans = parser.adapt_query(
        size,
        start,
        conf_mgr.elastic(mode=mode),
        elasticq,
        {
            "size": size,
            "sort": sort,
            "from_": start,
            "index": index,
            "_source_exclude": exclude,
            "version": settings["version"],
            "search_type": settings.get("search_type", None),
        },
    )

    _logger.debug("|requestquery| #hits ={n_hits}".format(n_hits=len(ans["hits"])))

    if settings.get("highlight", False):
        clean_highlight(ans)

    if settings.get("format") or settings.get("export"):
        formatmethod = "format" if "format" in settings else "export"
        toformat = settings.get(formatmethod)
        msg = "Unkown %s %s for mode %s" % (formatmethod, toformat, mode)
        format_posts = conf_mgr.extra_src(mode, formatmethod, helpers.notdefined(msg))
        format_posts(ans, conf_mgr.elastic(mode=mode), mode, index, toformat)

    # _logger.debug('|requestquery| returning ans={ans}'.format(ans=ans))
    return ans


def sortorder(settings, mode, querycommand):
    if not settings.get("sort"):
        if querycommand == "simple":
            # default: group by lexicon, then sort by score
            sort = conf_mgr.searchfield(mode, "sort_by")
        else:
            # default for extended query: exclude _score
            sort = [field for field in conf_mgr.searchfield(mode, "sort_by") if field != "_score"]
    else:
        sort = conf_mgr.searchfield(mode, "head_sort_field") + settings["sort"]

    return sort


def querycount(page=0):
    # TODO error if buckets is used here
    # TODO auth.validate_user is also done once in requestquery
    # but since we need the permitted dict, it is called
    # here as well
    user_is_authorized, permitted = auth.validate_user(mode="read")
    try:
        # TODO buckets should be gathered from some config
        stat_size = request.args.get("statsize", conf_mgr.app_config.MAX_PAGE)
        default = {"buckets": ["lexiconOrder", "lexiconName"], "size": stat_size}
        settings = parser.make_settings(permitted, default, user_is_authorized=user_is_authorized)
        q_ans = requestquery(page=page)

        # raise the size for the statistics call
        count_elasticq, more = parser.statistics(
            settings,
            order={"lexiconOrder": ("_key", "asc")},
            show_missing=False,
            force_size=stat_size,
        )
        mode = settings["mode"]
        es = conf_mgr.elastic(mode=mode)
        index, typ = conf_mgr.get_mode_index(mode)
        _logger.debug("|querycount| Will ask %s", count_elasticq)
        count_ans = es.search(
            index=index,
            body=count_elasticq,
            search_type="query_then_fetch",
            # raise the size for the statistics call
            size=25,  # stat_size
        )
        _logger.debug("ANNE: count_ans: %s\n", count_ans)
        distribution = count_ans["aggregations"]["q_statistics"]["lexiconOrder"]["buckets"]
    except errors.KarpException as e:  # pass on karp exceptions
        _logger.exception(e)
        raise

    except (elasticsearch.RequestError, elasticsearch.TransportError) as e:
        _logger.exception(e)
        raise errors.KarpElasticSearchError("ElasticSearch failure. Message: %s.\n" % e)

    except Exception as e:  # catch *all* exceptions
        # Remember that 'buckets' is not allowed here! %s"
        _logger.exception(e)
        raise errors.KarpQueryError("Could not parse data", debug_msg=e, query=request.query_string)
    return jsonify({"query": q_ans, "distribution": distribution})


def test():
    user_is_authorized, permitted = auth.validate_user(mode="read")
    try:
        # default
        settings = parser.make_settings(
            permitted, {"size": 25, "page": 0}, user_is_authorized=user_is_authorized
        )
        elasticq = parser.parse(settings)
    except QueryError as e:
        raise errors.KarpQueryError("Parse error", debug_msg=e, query=request.query_string)
    return jsonify({"elastic_json_query": elasticq})


def explain():
    user_is_authorized, permitted = auth.validate_user(mode="read")
    try:
        # default
        settings = parser.make_settings(
            permitted, {"size": 25, "page": 0}, user_is_authorized=user_is_authorized
        )
        elasticq = parser.parse(settings)
    except QueryError as e:
        raise errors.KarpQueryError("Parse error", debug_msg=e, query=request.query_string)
    es = conf_mgr.elastic(mode=settings["mode"])
    index, typ = conf_mgr.get_mode_index(settings["mode"])
    ex_ans = es.indices.validate_query(index=index, body=elasticq, explain=True)
    q_ans = requestquery(page=0)
    return jsonify({"elastic_json_query": elasticq, "ans": q_ans, "explain": ex_ans})


def minientry():
    """ Returns the counts and stats for the query """
    max_page = conf_mgr.app_config.MINIENTRY_PAGE
    user_is_authorized, permitted = auth.validate_user(mode="read")
    try:
        mode = parser.get_mode()
        default = {"show": conf_mgr.searchfield(mode, "minientry_fields"), "size": 25}
        settings = parser.make_settings(permitted, default, user_is_authorized=user_is_authorized)
        elasticq = parser.parse(settings)
        show = settings["show"]
        if not user_is_authorized:
            # show = show - exclude
            exclude = conf_mgr.searchfield(mode, "secret_fields")
            show = list(set(show).difference(exclude))

        sort = sortorder(settings, mode, settings.get("query_command", ""))
        start = settings["start"] if "start" in settings else 0
        es = conf_mgr.elastic(mode=settings["mode"])
        index, typ = conf_mgr.get_mode_index(settings["mode"])
        ans = parser.adapt_query(
            settings["size"],
            start,
            es,
            elasticq,
            {
                "index": index,
                "_source": show,
                "from_": start,
                "sort": sort,
                "size": min(settings["size"], max_page),
                "search_type": "dfs_query_then_fetch",
            },
        )
        if settings.get("highlight", False):
            clean_highlight(ans)

        return jsonify(ans)
    except AuthenticationError as e:
        _logger.exception(e)
        msg = e.message
        raise errors.KarpAuthenticationError(msg)
    except QueryError as e:
        raise errors.KarpQueryError(
            "Parse error, %s" % e.message, debug_msg=e, query=request.query_string
        )
    except errors.KarpException as e:  # pass on karp exceptions
        _logger.exception(e)
        raise
    except Exception as e:  # catch *all* exceptions
        _logger.exception(e)
        raise errors.KarpGeneralError("Unknown error", debug_msg=e, query=request.query_string)


def random():
    user_is_authorized, permitted = auth.validate_user(mode="read")
    try:
        mode = parser.get_mode()
        default = {"show": conf_mgr.searchfield(mode, "minientry_fields"), "size": 1}
        settings = parser.make_settings(permitted, default, user_is_authorized=user_is_authorized)
        elasticq = parser.random(settings)
        _logger.debug("random %s", elasticq)
        es = conf_mgr.elastic(mode=mode)
        index, typ = conf_mgr.get_mode_index(mode)
        es_q = {"index": index, "body": elasticq, "size": settings["size"]}
        if settings["show"]:
            show = settings["show"]
            if not user_is_authorized:
                # show = show - exclude
                exclude = conf_mgr.searchfield(mode, "secret_fields")
                show = list(set(show).difference(exclude))
            es_q["_source"] = show

        ans = es.search(**es_q)
        return jsonify(ans)
    except AuthenticationError as e:
        _logger.exception(e)
        msg = e.message
        raise errors.KarpAuthenticationError(msg)
    except errors.KarpException as e:  # pass on karp exceptions
        _logger.exception(e)
        raise
    except Exception as e:  # catch *all* exceptions
        _logger.exception(e)
        raise errors.KarpGeneralError("Unknown error", debug_msg=e, query=request.query_string)


def statistics():
    """ Returns the counts and stats for the query """
    user_is_authorized, permitted = auth.validate_user(mode="read")
    try:
        mode = parser.get_mode()
        default = {
            "buckets": conf_mgr.searchfield(mode, "statistics_buckets"),
            "size": 100,
            "cardinality": False,
        }
        settings = parser.make_settings(permitted, default, user_is_authorized=user_is_authorized)
        exclude = [] if user_is_authorized else conf_mgr.searchfield(mode, "secret_fields")

        elasticq, more = parser.statistics(settings, exclude=exclude)
        es = conf_mgr.elastic(mode=settings["mode"])
        index, typ = conf_mgr.get_mode_index(settings["mode"])
        is_more = check_bucketsize(more, settings["size"], index, es)

        # TODO allow more than 100 000 hits here?
        _logger.debug("stat body %s", elasticq)
        ans = es.search(index=index, body=elasticq, search_type="query_then_fetch", size=0)
        ans["is_more"] = is_more
        return jsonify(ans)
    except AuthenticationError as e:
        _logger.exception(e)
        msg = e.message
        raise errors.KarpAuthenticationError(msg)
    except errors.KarpException as e:  # pass on karp exceptions
        _logger.exception(e)
        raise
    except Exception as e:  # catch *all* exceptions
        _logger.exception(e)
        raise errors.KarpGeneralError("Unknown error", debug_msg=e, query=request.query_string)


def statlist():
    """ Returns the counts and stats for the query """
    user_is_authorized, permitted = auth.validate_user(mode="read")
    try:
        mode = parser.get_mode()
        _logger.debug("mode is %s", mode)
        default = {
            "buckets": conf_mgr.searchfield(mode, "statistics_buckets"),
            "size": 100,
            "cardinality": False,
        }
        settings = parser.make_settings(permitted, default, user_is_authorized=user_is_authorized)

        exclude = [] if user_is_authorized else conf_mgr.searchfield(mode, "secret_fields")
        elasticq, more = parser.statistics(settings, exclude=exclude, prefix="STAT_")
        es = conf_mgr.elastic(mode=settings["mode"])
        index, typ = conf_mgr.get_mode_index(settings["mode"])
        is_more = check_bucketsize(more, settings["size"], index, es)
        # TODO allow more than 100 000 hits here?
        size = settings["size"]
        ans = es.search(index=index, body=elasticq, search_type="query_then_fetch", size=0)
        tables = []
        for key, val in ans["aggregations"]["q_statistics"].items():
            if key.startswith("STAT_"):
                tables.extend(generate_table(val, []))
        # the length of tables might be longer than size, so truncate it
        # generating shorter tables is not faster than generating all of it
        # and then truncating
        if size:
            tables = tables[:size]
        return jsonify({"stat_table": tables, "is_more": is_more})

    except AuthenticationError as e:
        _logger.exception(e)
        msg = e.message
        raise errors.KarpAuthenticationError(msg)
    except errors.KarpException as e:  # pass on karp exceptions
        _logger.exception(e)
        raise
    except Exception as e:  # catch *all* exceptions
        # raise
        _logger.exception(e)
        raise errors.KarpGeneralError("Unknown error", debug_msg=e, query=request.query_string)


def check_bucketsize(bucket_sizes, size, index, es):
    is_more = {}
    for sizebucket, bucketname in bucket_sizes:
        countans = es.search(index=index, body=sizebucket, size=0, search_type="query_then_fetch")
        _logger.debug("countans %s", countans)
        bucketsize = countans["aggregations"]["more"]["value"]
        _logger.debug("bucketsize = (%s, %s) ", bucketsize, type(bucketsize))
        _logger.debug("size = (%s, %s) ", size, type(size))
        if int(bucketsize) > size:
            is_more[bucketname] = int(bucketsize)
    return is_more


def generate_table(dictionary, table):
    name = dictionary.get("key", "")
    end = True
    tables = []
    for key, val in dictionary.items():
        if key.startswith("STAT_"):
            end = False
            tables.extend(generate_table(val, table + [name]))
        if key == "buckets":
            end = False
            for bucket in val:
                tables.extend(generate_table(bucket, table))
    if end:
        count = dictionary.get("doc_count", 0)
        if count:
            tables.append(table + [name, count])

    return tables


def formatpost():
    """ Formats the posted data into wanted format
        The data should be a list
        Currently only working for saol
    """
    # get and parse data
    request.get_data()
    data = request.data
    try:
        data = json.loads(data)
    except ValueError as e:
        raise errors.KarpParsingError(str(e))

    # set all allowed lexicons (to avoid authentication exception
    user_is_authorized, permitted = auth.validate_user(mode="read")
    # find the wanted format
    settings = parser.make_settings(permitted, {"size": 25}, user_is_authorized=user_is_authorized)
    parser.parse_extra(settings)
    to_format = settings.get("format", "")
    mode = parser.get_mode()
    _logger.debug('mode "%s"', mode)
    index, typ = conf_mgr.get_mode_index(mode)

    if to_format:
        if not isinstance(data, list):
            data = [data]
        errmsg = "Unkown format %s for mode %s" % (settings["format"], mode)
        format_list = conf_mgr.extra_src(mode, "format_list", helpers.notdefined(errmsg))
        ok, html = format_list(data, conf_mgr.elastic(mode=mode), settings["format"], index)
        return jsonify({"all": len(data), "ok": ok, "data": html})

    else:
        raise errors.KarpQueryError("Unkown format %s" % to_format)


def autocomplete():
    """ Returns lemgrams matching the query text.
        Each mode specifies in the configs which fields that should be
        considered.
        The parameter 'q' or 'query' is used when only one word form is to be
        processed.
        The parameter 'multi' is used when multiple word forms should be
        processed.
        The format of result depends on which flag that is set.
    """
    user_is_authorized, permitted = auth.validate_user(mode="read")
    # query = request.query_string
    try:
        settings = parser.make_settings(
            permitted, {"size": 1000}, user_is_authorized=user_is_authorized
        )
        mode = parser.get_mode()
        resource = parser.parse_extra(settings)
        if "q" in request.args or "query" in request.args:
            qs = [request.args.get("q", "") or request.args.get("query", "")]
            _logger.debug("qs is %s", qs)
            multi = False
        else:
            # check if there are multiple words forms to complete
            qs = settings.get("multi", [])
            _logger.debug("qs %s", qs)
            multi = True

        filters = []  # filters will be put here
        if not settings.get("user_is_authorized", False):
            filter_unauth_user = conf_mgr.filter_for_unauth_user(mode)
            if filter_unauth_user is not None:
                filters.append({"term": filter_unauth_user})
        # use utf8, escape '"'
        qs = [re.sub('"', '\\"', q) for q in qs]

        headboost = conf_mgr.searchfield(mode, "boosts")[0]
        res = {}
        single_response = {}
        # if multi is not true, only one iteration of this loop will be done
        for q in qs:
            boost = {"term": {headboost: {"boost": "500", "value": q}}}

            autocompleteq = conf_mgr.extra_src(mode, "autocomplete", autocompletequery)
            exp = autocompleteq(mode, boost, q)
            autocomplete_field = conf_mgr.searchonefield(mode, "autocomplete_field")
            autocomplete_fields = conf_mgr.searchfield(mode, "autocomplete_field")
            fields = {"exists": {"field": autocomplete_field}}
            # last argument is the 'fields' used for highlightning
            elasticq = parser.search([exp, fields, resource], filters, "", usefilter=True)
            _logger.debug("Will send %s", elasticq)

            es = conf_mgr.elastic(mode=mode)
            _logger.debug("_source: %s", autocomplete_field)
            _logger.debug(elasticq)
            index, typ = conf_mgr.get_mode_index(mode)
            single_response = parser.adapt_query(
                settings["size"],
                0,
                es,
                elasticq,
                {"size": settings["size"], "index": index, "_source": autocomplete_fields,},
            )
            # save the results for multi
            res[q] = single_response
        if multi:
            return jsonify(res)
        else:
            # single querys: only return the latest answer
            return jsonify(single_response)
    except AuthenticationError as e:
        _logger.exception(e)
        msg = e.message
        raise errors.KarpAuthenticationError(msg)
    except errors.KarpException as e:  # pass on karp exceptions
        _logger.exception(e)
        raise
    except Exception as e:  # catch *all* exceptions
        _logger.exception(e)
        raise errors.KarpGeneralError("Unknown error", debug_msg=e, query=request.query_string)


# standard autocomplete
def autocompletequery(mode, boost, q):
    """ Constructs an autocompletion query, searching for lemgrams starting
        with 'text'
        Returns a query object to be sent to elastic search
    """
    # other modes: don't care about msd
    look_in = [boost]
    for boost_field in conf_mgr.searchfield(mode, "boosts"):
        look_in.append({"match_phrase": {boost_field: q}})

    exp = {"bool": {"should": look_in}}

    return exp


def clean_highlight(ans):
    stop_offset = 9  # The number of extra tokens added by the <em> tags
    for n, hit in enumerate(ans.get("hits", {}).get("hits", [])):
        # _logger.debug('CLEAN hit %s\n\n\n' % hit)
        for field, texts in hit.get("highlight", {}).items():
            # _logger.debug('CLEAN texts %s: %s' % (field, texts))
            if field == "lexiconName":
                del ans["hits"]["hits"][n]["highlight"][field]
            else:
                newtexts = chain(*[re.finditer("<em>(.*?)</em>", t) for t in texts])
                spans = []
                for new in newtexts:
                    spans.append((new.group(1), new.span()[0], new.span()[1] - stop_offset))
                ans["hits"]["hits"][n]["highlight"][field] = spans
    ans["hits"]["highlight"] = "ON"


def lexiconorder():
    orderlist = {}
    for name, val in conf_mgr.lexicons.items():
        orderlist[name] = val.get("order", "-1")
    return jsonify(orderlist)


def modeinfo(mode):
    return jsonify(conf_mgr.fields.get(mode, {}))


def lexiconinfo(lexicon):
    return jsonify(conf_mgr.fields.get(conf_mgr.get_lexicon_mode(lexicon), {}))


# For debugging
def testquery():
    """ Returns the query expressed in elastics search api """
    user_is_authorized, permitted = auth.validate_user(mode="read")
    try:
        # default
        settings = parser.make_settings(
            permitted, {"size": 25, "page": 0}, user_is_authorized=user_is_authorized
        )
        elasticq = parser.parse(settings)
        mode = settings["mode"]
        if not settings.get("sort", ""):
            # default: group by lexicon, then sort by score
            sort = conf_mgr.searchfield(mode, "sort_by")
        else:
            sort = settings["sort"]
        start = settings["start"] if "start" in settings else settings["page"] * settings["size"]
        elasticq = parser.parse(settings)
        return json.dumps(elasticq) + json.dumps(
            {"sort": sort, "_from": start, "size": settings["size"], "version": "true"}
        )
    except Exception as e:  # catch *all* exceptions
        # TODO only catch relevant exceptions
        _logger.exception(e)
        raise errors.KarpGeneralError(e, request.query_string)


def get_context(lexicon):
    """ Find and return the alphabetically (or similar, as specified for the
    lexicon) context of a word/entry.
    """
    user_is_authorized, permitted = auth.validate_user(mode="read")
    if lexicon not in permitted:
        raise errors.KarpAuthenticationError(
            "You are not allowed to search the " "lexicon %s" % lexicon
        )
    # make default settings
    settings = parser.make_settings(
        permitted, {"size": 10, "resource": lexicon}, user_is_authorized=user_is_authorized
    )
    # parse parameter settings
    parser.parse_extra(settings)

    # set searching configurations
    mode = conf_mgr.get_lexicon_mode(lexicon)
    settings["mode"] = mode
    es = conf_mgr.elastic(mode=mode)
    index, typ = conf_mgr.get_mode_index(mode)

    # get the sort_by list (eg. ['baseform.sort', 'lemmaid.search'])
    # leave out lexiconOrder and _score
    sortfieldnames = [
        field
        for field in conf_mgr.searchconf(mode, "sort_by")
        if field not in ["_score", "lexiconOrder"]
    ]
    _logger.debug("|get_context| sortfieldnames = %s", sortfieldnames)
    # get the sort field paths (eg. ['FormRep.baseform.raw', 'lemmaid.raw'])
    # Used for sorting.
    sortfield = sum([conf_mgr.lookup_multiple(f, mode) for f in sortfieldnames], [])
    # get the field name of the head sort field. Used for searching
    sortfieldname = sortfieldnames[0]

    filters = []  # filters will be put here
    print(f"mode = {mode}")
    if not settings.get("user_is_authorized", False):
        filter_unauth_user = conf_mgr.filter_for_unauth_user(mode)
        print(f"filter_unauth_user = {filter_unauth_user}")
        if filter_unauth_user is not None:
            filters.append({"term": filter_unauth_user})
        print(f"filters = {filters}")

    # find the center entry (by its id)
    if "center" in settings:
        center_id = settings["center"]
        id_query = {"term": {"_id": center_id}}
        if filters:
            body_query = {"query": {"bool": {"must": id_query, "filter": filters}}}
        else:
            body_query = {"query": id_query}

        lexstart = es.search(
            index=index,
            doc_type=typ,
            size=1,
            body=body_query,
            sort=["%s:asc" % f for f in sortfield],
        )
    # if no center id is given, pick the first entry of the lexicon
    else:
        exps = []
        parser.parse_ext("and|resource|equals|%s" % lexicon, exps, [], mode)
        center_q = parser.search(exps, filters, [], usefilter=True, constant_score=False)
        center_q = {"query": center_q}
        lexstart = es.search(
            index=index,
            doc_type=typ,
            size=1,
            body=center_q,
            sort=["%s:asc" % f for f in sortfield],
        )
        center_id = lexstart["hits"]["hits"][0]["_id"]

        # lexstart = es.search(index=index, doc_type=typ, size=1,
        #                      sort=['%s:asc' % f for f in sortfield])
        # center_id = lexstart['hits']['hits'][0]['_id']

    if not lexstart["hits"]["hits"]:
        _logger.error("No center found %s, %s", center_id, lexstart)
        raise errors.KarpElasticSearchError("Could not find entry %s" % center_id)

    centerentry = lexstart["hits"]["hits"][0]
    _logger.debug("center %s, %s", centerentry, centerentry["_id"])
    origentry_sort = [key for key in centerentry["sort"] if key is not None][0]
    # TODO what to do if the sort key is not in the lexicon? as below?
    # origentry_sort = centerentry['sort'][0]
    # TODO test this!
    if not isinstance(origentry_sort, str):
        origentry_sort = str(origentry_sort)
    sortvalue = util.escape_control(origentry_sort)
    _logger.debug("Orig entry escaped key %s", sortvalue)

    # Construct queries to ES
    exps = []
    # the query string from the user
    querystring = settings.get("q", "")
    parser.parse_ext("and|resource|equals|%s" % lexicon, exps, [], mode)
    if querystring:
        if querystring.startswith("simple"):
            querystring = "and|anything|equals|%s" % querystring.split("|")[-1]
        else:
            querystring = re.sub(r"extended\|\|", "", querystring)
        parser.parse_ext(querystring, exps, [], mode)

    preexps = copy.deepcopy(exps)  # deep copy for the pre-query
    hits_post = get_pre_post(
        exps,
        center_id,
        sortfield,
        sortfieldname,
        sortvalue,
        mode,
        settings["size"],
        es,
        index,
        place="post",
        filters=filters,
    )
    hits_pre = get_pre_post(
        preexps,
        center_id,
        sortfield,
        sortfieldname,
        sortvalue,
        mode,
        settings["size"],
        es,
        index,
        place="pre",
        filters=filters,
    )
    return jsonify(
        {
            "pre": hits_pre[: settings["size"]],
            "post": hits_post[: settings["size"]],
            "center": centerentry,
        }
    )


def get_pre_post(
    exps,
    center_id,
    sortfield,
    sortfieldname,
    sortvalue,
    mode,
    size: int,
    es,
    index,
    place="post",
    filters=None,
):
    op = {"post": {"op": "gte", "sort": "asc"}, "pre": {"op": "lte", "sort": "desc"}}
    parser.parse_ext("and|%s|%s|%s" % (sortfieldname, op[place]["op"], sortvalue), exps, [], mode)
    elasticq_q = parser.search(
        exps, filters if filters else [], [], usefilter=False, constant_score=True
    )

    # +1 to compensate for the word itself being in the context
    size = size + 1
    show = conf_mgr.searchfield(mode, "minientry_fields")
    for _i, _v in enumerate(show):
        if _v == "Corpus_unit_id.raw":
            show[_i] = "Corpus_unit_id"
    _logger.debug("searching.py:get_pre_post show = %s", show)
    # TODO size*3 (magic number) because many entries may have the same sort
    # value (eg homographs in saol)
    ans = parser.adapt_query(
        size * 3,
        0,
        es,
        elasticq_q,
        {
            "size": size * 3,
            "from_": 0,
            "sort": ["%s:%s" % (f, op[place]["sort"]) for f in sortfield],
            "index": index,
            "_source": show,
        },
    )
    hits = ans.get("hits", {}).get("hits", [])
    return go_to_sortkey(hits, center_id)


def go_to_sortkey(hits, center_id):
    """ Step through the lists until the center entry is passed
        The center is at the beginning of the list, or after its homonymns
        (entries with the same primary sort key). Pass these.
        Stop counting when the center_id is passed.
    """
    n = 1
    stopped = True
    for hit in hits:
        if center_id == hit["_id"]:
            stopped = True
            break
        n += 1
    # if the center, for some reson, is not among the hits, show all hits
    n = 0 if not stopped else n
    return hits[n:]


def export(lexicon):
    # TODO can user with only read permissions export all the lexicon?
    # (eg saol)
    user_is_authorized, permitted = auth.validate_user(mode="read")
    if not user_is_authorized or lexicon not in permitted:
        raise errors.KarpAuthenticationError(
            "You are not allowed to export the lexicon '%s'" % lexicon
        )

    settings = parser.make_settings(
        permitted, {"size": -1, "resource": lexicon}, user_is_authorized=user_is_authorized
    )
    parser.parse_extra(settings)
    date = settings.get("date", "")
    mode = settings.get("mode", "")
    if date:
        import dateutil.parser as dateP
        from datetime import datetime

        # parse the date as inclusive (including the whole selected day)
        date = dateP.parse(
            date, dateP.parserinfo(yearfirst=True), default=datetime(1999, 0o1, 0o1, 23, 59),
        )

    # to_keep = {}
    # engine, db_entry = db.get_engine(lexicon, echo=False)
    _logger.debug("exporting entries from %s ", lexicon)
    entries = db.get_entries_to_keep_gen(lexicon, to_date=date)
    #    for entry in db.dbselect(
    #        lexicon, engine=engine, db_entry=db_entry, max_hits=-1, to_date=date
    #    ):
    #        _id = entry["id"]
    #        if _id in to_keep:
    #            last = to_keep[_id]["date"]
    #            if last < entry["date"]:
    #                to_keep[_id] = entry
    #        else:
    #            to_keep[_id] = entry
    #
    #    ans = [val["doc"] for val in list(to_keep.values()) if val["status"] != "removed"]
    size = settings["size"]
    if isinstance(size, int) and size >= 0:
        entries = itertools.islice(entries, size)
        # ans = itertools.islice(ans, size)
        # ans = ans[:size]

    # _logger.debug("exporting %s entries", len(ans))
    if settings.get("format", ""):
        toformat = settings.get("format")
        msg = "Unkown %s %s for mode %s" % ("format", toformat, mode)
        format_posts = conf_mgr.extra_src(mode, "exportformat", helpers.notdefined(msg))
        lmf, err = format_posts(entries, lexicon, mode, toformat)
        return Response(lmf, mimetype="text/xml")

    else:
        # divsize = 5000
        # if len(ans) < divsize:
        #     _logger.debug("simply sending entries")
        #     return jsonify({lexicon: ans})

        def gen():
            # start, stop = 0, divsize
            yield '{"%s": [' % lexicon
            for i, obj in enumerate(entries):
                if i > 0:
                    yield ","
                yield json.dumps(obj)
            #            while start < len(ans):
            #                _logger.debug("chunk %s - %s" % (start, stop))
            #                if start > 1:
            #                    yield ","
            #                yield ",".join([json.dumps(obj) for obj in ans[start:stop]])
            #                # yield json.dumps(ans[start:stop])
            #                start = stop
            #                stop += divsize
            yield "]}"

        _logger.debug("streaming entries")
        return Response(stream_with_context(gen()))



"""Handle lexicon and modes.
"""
import datetime
import json
import logging
import os
import sys
from typing import IO, Optional, List, Union

from elasticsearch import helpers as es_helpers
from elasticsearch import exceptions as esExceptions

from sb_json_tools import json_iter

# import elasticsearch_dsl as es_dsl

import karp5.dbhandler.dbhandler as db
from karp5.config import mgr as conf_mgr
from karp5 import document
from karp5.server.translator import bulkify


_logger = logging.getLogger("karp5")

# ==============================================
# helpers
# ==============================================


def make_indexname(index, suffix=None):
    if not suffix:
        suffix = datetime.datetime.now().strftime("%Y%m%d-%H%M")
    return "{index}_{suffix}".format(index=index, suffix=suffix)


def update_source_field(mode, doc):
    """ Apply doc_to_es to doc. """
    doc["_source"] = document.doc_to_es(doc["_source"], mode, "update")
    return doc


def index_create(mode, index):
    """ Create a new Elasticsearch index. """

    es = conf_mgr.elastic(mode)
    data = conf_mgr.get_mapping(mode)
    try:
        _logger.debug("Creating index '%s' for mode '%s' with es=%s", index, mode, es)
        ans = es.indices.create(  # pylint: disable=unexpected-keyword-arg
            index=index, body=data, request_timeout=30
        )
    except esExceptions.TransportError as e:
        _logger.exception(e)
        raise Exception("Could not create index")
    return ans


def index_exists(mode: str, index: Union[str, List[str]]) -> bool:
    """Check if index exists for mode.

    Arguments:
        mode {str} -- the mode to query from
        index {str} -- the index to to query for

    Returns:
        bool -- wheter the index exists or not
    """
    es = conf_mgr.elastic(mode)
    ans = es.indices.exists(index=index)
    print(f"index_exists: and = {ans}")
    return ans


def upload(
    informat, name, order, data, elastic, index, typ, sql=False, verbose=True, with_id=False,
):
    """Uploads the data to elastic and the database
        sql      if True,  the data will be stored in the SQL database as well
                 as ElasticSearch
                 if False, the data will only be stored in ElasticSearch
        informat can either be xml  - lmf
                               json - a single json object or a list of objects
                               bulk - a list of json objects annotated with
                                      index and type information, as accepted
                                      by ElasticSearch

    Arguments:
        informat {[type]} -- [description]
        name {[type]} -- [description]
        order {[type]} -- [description]
        data {[type]} -- [description]
        elastic {[type]} -- [description]
        index {[type]} -- [description]
        typ {[type]} -- [description]

    Keyword Arguments:
        sql {bool} -- [description] (default: {False})
        verbose {bool} -- [description] (default: {True})
        with_id {bool} -- [description] (default: {False})

    Raises:
        Exception: [description]
        Exception: [description]
        Exception: [description]
        Exception: [description]
        Exception: [description]
        Exception: [description]
        Exception: [description]

    Returns:
        [type] -- [description]
    """
    try:
        # The actual parsing
        data = parse_upload(informat, name, order, data, index, typ, with_id=with_id)
    except Exception:
        print("Error while reading data from %s" % name)
        raise

    ok = 0
    if sql:
        # stream entries one by one to elastic, then update sql db
        # streaming_bulk will notify us at once when an entry fails
        sql_bulk = []
        for res in es_helpers.streaming_bulk(elastic, data, request_timeout=60):
            # res is a tuple, res[0]==True
            ansname = "index"  # if with_id else 'create' -- changed from ES2
            _id = res[1].get(ansname).get("_id")
            data_doc = data[ok].get("_source")
            if not isinstance(data_doc, dict):
                data_doc = json.loads(data_doc)
            sql_doc = document.doc_to_sql(data_doc, data_doc["lexiconName"], "bulk")
            sql_bulk.append(
                (
                    _id,
                    json.dumps(sql_doc),
                    "admin",
                    "entry automatically added or reloaded",
                    name,
                    "imported",
                )
            )
            ok += 1
        db_loaded, db_error = db.update_bulk(name, sql_bulk)
        if db_error:
            _logger.error("Got DB error: %s", db_error)
            raise Exception(db_error)
        ok += db_loaded
    else:
        # upload all at once to elastic
        ok, err = es_helpers.bulk(elastic, data, request_timeout=60)
        if err:
            msg = "Error during upload. %s documents successfully uploaded. \
                   Message: %s.\n"
            raise Exception(msg % (ok, "\n".join(err)))
    if not ok:
        _logger.warning("No data. 0 documents uploaded.")
        raise Exception("No data")
    if verbose:
        print("Ok. %s documents uploaded\n" % ok)


def parse_upload(informat, lexname, lexOrder, data, index, typ, with_id=False):
    """Parse the query and the post and put it into the desired json format

    Arguments:
        informat {[type]} -- [description]
        lexname {[type]} -- [description]
        lexOrder {[type]} -- [description]
        data {[type]} -- [description]
        index {[type]} -- [description]
        typ {[type]} -- [description]

    Keyword Arguments:
        with_id {bool} -- [description] (default: {False})

    Returns:
        [type] -- [description]
    """

    bulk_info = {"index": index, "type": typ}
    out_data = 0, [], []

    if informat == "json":
        out_data = bulkify.bulkify(data, bulk_info, with_id=with_id)
    else:
        raise "Don't know how to parse %s" % informat

    return out_data


# def get_entries_to_keep_from_sql(lexicons):
#     to_keep = {}
#     if isinstance(lexicons, six.text_type):
#         lexicons = [lexicons]

#     for lex in lexicons:
#         engine, db_entry = db.get_engine(lex, echo=False)
#         for entry in db.dbselect(lex, engine=engine, db_entry=db_entry, max_hits=-1):
#             _id = entry["id"]
#             if _id:  # don't add things without id, they are errors
#                 if _id in to_keep:
#                     last = to_keep[_id]["date"]
#                     if last < entry["date"]:
#                         _logger.debug("|get_entries_to_keep_from_sql| Update entry.")
#                         to_keep[_id] = entry
#                 else:
#                     to_keep[_id] = entry
#             else:
#                 _logger.warning("|sql no id| Found entry without id:")
#                 _logger.warning("|sql no id| %s", entry)
#     # _logger.debug("to_keep = %s", to_keep)
#     return to_keep


def recover(alias, suffix, lexicon, create_new=True) -> bool:
    """ Recovers the data to ES, uses SQL as the trusted base version.
        Find the last version of every SQL entry and send this to ES.
    """

    # if not lexicon:
    #     lexicon = conf.keys()
    index = make_indexname(alias, suffix)
    index_type = conf_mgr.get_mode_type(alias)
    print("Save %s to %s" % (lexicon or "all", index))

    es = conf_mgr.elastic(alias)
    if create_new:
        # Create the index
        index_create(alias, index)

    # to_keep = get_entries_to_keep_from_sql(lexicon)
    # print(len(to_keep), "entries to keep")
    entries = db.get_entries_to_keep_gen(lexicon)

    data = bulkify.bulkify_from_sql(entries, index, index_type)
    ok, err = es_helpers.bulk(es, data, request_timeout=30)
    if err:
        msg = "Error during upload. %s documents successfully uploaded. \
               Message: %s.\n"
        raise Exception(msg % (ok, "\n".join(err)))
    print("recovery done")
    return True


# def recover_add(index, suffix, lexicon):
#     # xTODO test this
#     """ Recovers the data to ES, uses SQL as the trusted base version.
#         Find the last version of every SQL entry and send this to ES.
#         Adds the specified lexicons to an existing index
#     """

#     es = conf_mgr.elastic(index)
#     print("Save %s to %s" % (lexicon, index))

#     to_keep = get_entries_to_keep_from_sql(lexicon)
#     print(len(to_keep), "entries to keep")

#     data = bulkify.bulkify_sql(to_keep, bulk_info={"index": index})
#     ok, err = es_helpers.bulk(es, data, request_timeout=30)
#     if err:
#         msg = "Error during upload. %s documents successfully uploaded. \
#                Message: %s.\n"
#         raise Exception(msg % (ok, "\n".join(err)))
#     print("recovery done")


def printlatestversion(lexicon: str, with_id: bool = False, fp: Optional[IO[str]] = None):
    """Dump the latest entries for a lexicon (or mode?).

       If with_id=True, then the results can be imported from cli.
    Arguments:
        lexicon -- the lexicon to print

    Keyword Arguments:
        debug -- whether to print debug information (default: {True})
        with_id -- if True, prints the entries as in ElasticSearch (default: {False})
        fp -- file handle to write to, if None write to stdout (default: {None})
    """
    if fp is None:
        fp = sys.stdout

    # to_keep = get_entries_to_keep_from_sql(lexicon)
    entries = db.get_entries_to_keep_gen(lexicon)

    if with_id:
        entries = (
            {"_id": entry["id"], "_source": document.doc_to_es(entry["doc"], lexicon, "bulk")}
            for entry in entries
        )
    else:
        entries = (entry["doc"] for entry in entries)

    json_iter.dump(entries, fp)


def publish_mode(mode, suffix):
    index = make_indexname(mode, suffix)
    add_actions = []
    rem_actions = []
    for alias in conf_mgr.get_modes_that_include_mode(mode):
        add_actions.append('{"add" : {"index": "%s", "alias": "%s"}}' % (index, alias))
        rem_actions.append('{"remove": {"index":"%s_*", "alias":"%s"}}' % (mode, alias))
    print("remove {}".format(rem_actions))
    print("add {}".format(add_actions))

    es = conf_mgr.elastic(mode)

    try:
        print("remove old aliases")
        es.indices.update_aliases(  # pylint: disable=unexpected-keyword-arg
            '{"actions" : [%s]}' % ",".join(rem_actions), request_timeout=30
        )  # pylint: disable=unexpected-keyword-arg
    except esExceptions.ElasticsearchException:
        print("No previous aliased indices, could not do remove any")
        print(rem_actions)

    return es.indices.update_aliases(  # pylint: disable=unexpected-keyword-arg
        '{"actions" : [%s]}' % ",".join(add_actions), request_timeout=30
    )


# TODO Is this used?
# def publish_group(group, suffix):
#     # xTODO for some reason, this sometimes removes siblings
#     # of the group from an mode. Eg. publish_group(saldo)
#     # may lead to 'saldogroup' not containing 'external'.
#     es = conf_mgr.elastic(group)
#     print(group, suffix)
#     if not conf_mgr.modes.get(group)["is_index"]:
#         for subgroup in conf_mgr.modes.get(group)["groups"]:
#             publish_group(subgroup, suffix)

#     else:
#         name = make_indexname(group, suffix)
#         print("Publish %s as %s" % (name, group))
#         add_actions = []
#         rem_actions = []
#         for parent in conf_mgr.get_modes_that_include_mode(group):
#             add_actions.append('{"add" : {"index": "%s", "alias": "%s"}}' % (name, parent))
#             rem_actions.append('{"remove": {"index":"%s_*", "alias":"%s"}}' % (group, parent))

#         print("remove", rem_actions)
#         print("add", add_actions)
#         try:
#             print("remove old aliases")
#             es.indices.update_aliases(
#                 '{"actions" : [%s]}' % ",".join(rem_actions), request_timeout=30
#             )
#         except Exception:
#             print("No previous aliased indices, could not do remove any")
#             print(rem_actions)
#         return es.indices.update_aliases(
#             '{"actions" : [%s]}' % ",".join(add_actions), request_timeout=30
#         )


def create_empty_index(mode, suffix):
    return index_create(mode, make_indexname(mode, suffix))


def create_mode(alias, suffix, with_id=False, data_dir=None):
    """[summary]

    Arguments:
        alias {[type]} -- [description]
        suffix {[type]} -- [description]

    Keyword Arguments:
        with_id {bool} -- [description] (default: {False})
        data_dir -- Path to load lexicons from. (default: {None})
    """
    es = conf_mgr.elastic(alias)
    if conf_mgr.modes[alias]["is_index"]:
        to_create = [alias]
    else:
        to_create = conf_mgr.modes.get(alias)["groups"]

    typ = conf_mgr.modes[alias]["type"]
    for index in to_create:
        newname = make_indexname(index, suffix)
        index_create(alias, newname)
        try:
            lexicons = conf_mgr.get_lexiconlist(index)
            load(lexicons, newname, typ, es, with_id=with_id, data_dir=data_dir)
        except Exception as e:
            # delete the index if things did not go well
            ans = es.indices.delete(newname)
            _logger.error(
                "Any documentes uploaded to ES index %s are removed. ans = '%s'", newname, ans,
            )
            _logger.error("If data was uploaded to SQL you will have to remove it manually.")
            _logger.exception(e)
            raise


def add_lexicon(lexicon, suffix, filename=None, with_id=False):
    mode = conf_mgr.get_lexicon_mode(lexicon)
    es = conf_mgr.elastic(mode)
    indexname = make_indexname(mode, suffix)
    _type = conf_mgr.modes[mode]["type"]
    _logger.debug(
        "add_lexicon(lexicon=%s suffix=%s filename=%s with_id=%s)",
        lexicon,
        suffix,
        filename,
        with_id,
    )
    _logger.debug(" -> mode=%s indexname=%s es=%s", mode, indexname, es)
    if not es.indices.exists(indexname):
        try:
            ans = index_create(mode, indexname)
        except Exception:
            print("Could not create index. Check if it needs manual removal")
            raise
    try:
        if filename is None:
            filename = conf_mgr.default_filename(lexicon)
        with open(filename, "r") as fp:
            data = fp.read()
        sql = conf_mgr.get_mode_sql(mode)
        upload("json", lexicon, "", data, es, indexname, _type, sql=sql, with_id=with_id)
        # do this last to get the avoid version conflicts (in case some user is
        # updating the index during this process)
        # reindex_alias(mode, suffix, create_index=False)

    except Exception as e:
        # delete the index if things did not go well
        ans = es.indices.delete(indexname)
        # print(ans)
        _logger.error("Any documentes uploaded to ES index %s are removed. ans=%s", indexname, ans)
        _logger.error("If data was uploaded to SQL you will have to remove it manually.")
        _logger.exception(e)
        raise
    return indexname


def internalize_lexicon(mode, to_add):
    """ Reads all entries from the lexicons specified in to_add
        and puts them in sql. They may then be edited.
        The ES index is not effected and this operation may hence
        be run at any time
    """
    ok = 0
    es = conf_mgr.elastic(mode)
    sql_bulk = None
    for lex in to_add:
        print("Internalize", lex)
        # Go through each lexicon separately
        query = {"query": {"term": {"lexiconName": lex}}}
        # scan and scroll
        ans = es_helpers.scan(
            es,
            query=query,
            scroll="3m",
            raise_on_error=True,
            preserve_order=False,
            index=mode,
            request_timeout=30,
        )
        sql_bulk = []
        for hit in ans:  # ans is an iterator of objects from in hits.hits
            _id = hit.get("_id")
            source = hit.get("_source")
            if not isinstance(source, dict):
                source = json.loads(source)
            sql_doc = document.doc_to_sql(source, lex, "bulk")
            sql_bulk.append(
                (
                    _id,
                    json.dumps(sql_doc),
                    "admin",
                    "entry automatically added or reloaded",
                    lex,
                    "imported",
                )
            )

        db_loaded, db_error = db.update_bulk(lex, sql_bulk)
        if db_error:
            raise Exception(db_error)
        ok += db_loaded

    if sql_bulk:
        print("will load %s entries, starting with %s" % (len(sql_bulk), sql_bulk[0]))
    if not ok:
        _logger.warning("No data. 0 documents uploaded.")
        raise Exception("No data")
    print("Ok. %s documents loaded to sql\n" % ok)


def load(to_upload, index, typ, es, with_id: bool = False, data_dir: str = None):
    """[summary]

    Arguments:
        to_upload {[type]} -- [description]
        index {[type]} -- [description]
        typ {[type]} -- [description]
        es {[type]} -- [description]

    Keyword Arguments:
        with_id {bool} -- [description] (default: {False})
        data_dir {[type]} -- [description] (default: {None})
    """
    print("Upload to %s" % index, ",".join(to_upload))
    try:
        for lexicon, info in conf_mgr.lexicons.items():
            if lexicon in to_upload or not to_upload:
                default = conf_mgr.lexicons.get("default", {})
                file_format = info.get("format", default.get("format", "json"))
                if data_dir is not None:
                    filename = os.path.join(data_dir)
                filename = conf_mgr.default_filename(lexicon)
                with open(filename, "r") as fp:
                    _logger.info("reading file '%s'", filename)
                    data = fp.read()
                sql = conf_mgr.modes[info["mode"]]["sql"]
                _logger.info("Upload %s. To sql? %s", lexicon, sql)
                upload(
                    file_format,
                    lexicon,
                    info["order"],
                    data,
                    es,
                    index,
                    typ,
                    sql=sql,
                    with_id=with_id,
                )
                _logger.info("Upload of %s finished.", lexicon)
    except Exception:
        _logger.error(
            """Error during upload.
                 Check you're data bases for partially uploaded data."""
        )
        raise


def apply_filter(it, filter_func, field=None):
    """
    Apply the given func to every object in it.

    :param it: The iterable.
    :param filter_func: The func to apply. Must return an iterable.
    :param field: str
    """
    if not field:
        field = "_source"
    filtered = []
    for i in it:
        for x in filter_func(i[field], i["_id"]):
            filtered.append({field: x})
    return filtered


def copy_alias_to_new_index(
    source_mode, target_mode, target_suffix, filter_func=None, create_index=True, query=None,
):
    es_source = conf_mgr.elastic(source_mode)
    es_target = conf_mgr.elastic(target_mode)
    target_index = make_indexname(target_mode, target_suffix)
    _logger.debug("Copying from '%s' to '%s'", source_mode, target_index)
    if create_index:
        index_create(target_mode, target_index)

    source_docs = es_helpers.scan(es_source, index=source_mode, query=query)

    if filter_func:
        source_docs = apply_filter(source_docs, filter_func)

    target_type = conf_mgr.get_mode_type(target_mode)

    def update_doc(doc):
        """ Apply doc_to_es to doc. """
        doc["_source"] = document.doc_to_es(doc["_source"], target_mode, "update")
        doc["_index"] = target_index
        doc["_type"] = target_type
        return doc

    update_docs = (update_doc(doc) for doc in source_docs)
    success = 0
    errors = []
    for ok, item in es_helpers.streaming_bulk(es_target, update_docs, index=target_index):
        if not ok:
            errors.append(item)
        else:
            success += 1
    if len(errors) == 0:
        print("Done! Reindexed {} entries".format(success))
        return True, None
    else:
        print("Something went wrong!")
        print("  - Successfully reindexed: {}".format(success))
        print("  - Failed to reindex: {}".format(len(errors)))
        print("This are the failing entries:")
        print(errors)
        return False, errors


def reindex(alias, source_suffix, target_suffix, create_index=True):
    source_index = make_indexname(alias, source_suffix)
    target_index = make_indexname(alias, target_suffix)
    return reindex_help(alias, source_index, target_index, create_index=create_index)


def reindex_alias(alias, target_suffix, create_index=True):
    target_index = make_indexname(alias, target_suffix)
    return reindex_help(alias, alias, target_index, create_index=create_index)


def reindex_help(alias, source_index, target_index, create_index=True):
    print("Reindex from %s to %s" % (source_index, target_index))
    es = conf_mgr.elastic(alias)
    if create_index:
        index_create(alias, target_index)

    source_docs = es_helpers.scan(es, size=10000, index=source_index, raise_on_error=True)

    def update_doc(doc):
        """ Apply doc_to_es to doc. """
        doc["_source"] = document.doc_to_es(doc["_source"], alias, "update")
        doc["_index"] = target_index
        return doc

    update_docs = (update_doc(doc) for doc in source_docs)
    success, failed, total = 0, 0, 0
    errors = []
    for ok, item in es_helpers.streaming_bulk(es, update_docs, index=target_index):
        if not ok:
            failed += 1
            errors.append(item)
        else:
            success += 1
            print("ok = {},item = {}".format(ok, item))
        total += 1
    if success == total:
        _logger.info("Done! Reindexed %s entries", total)
        return True
    else:
        _logger.warning("Something went wrong!")
        _logger.warning("  - Successfully reindexed: %s", success)
        _logger.warning("  - Failed to reindex: %s", failed)
        _logger.warning("This are the failing entries:")
        _logger.warning(errors)
        return False


# TODO Is this used?
# def publish_all(suffix):
#     for alias, aliasconf in conf_mgr.modes.items():
#         # Only publish if it is a group, meta-aliases will point to the correct
#         # subaliases anyway.
#         if aliasconf["is_index"]:
#             publish_group(alias, suffix)


# TODO Is this used?
# def make_structure():
#     add_actions = []
#     # xTODO does not work, what is confelastic?
#     for alias, aliasconf in conf_mgr.modes.items():
#         # Only publish if it is a group, meta-aliases will point to the correct
#         # subaliases anyway.
#         es = conf_mgr.elastic(alias)
#         if not aliasconf.get("is_index"):
#             # if it is a mode (not just an index), remove the old pointers
#             add_actions.append('{"remove": {"index":"*", "alias":"%s"}}' % alias)
#         for group in aliasconf.get("groups", []):
#             add_actions.append('{"add" : {"index": "%s", "alias": "%s"}}' % (group, alias))

#     return es.indices.update_aliases(
#         '{"actions" : [%s]}' % ",".join(add_actions), request_timeout=30
#     )


def delete_all():
    # delete all indices
    for alias, _ in conf_mgr.modes.items():
        es = conf_mgr.elastic(alias)
        try:
            es.indices.delete("*")
        except esExceptions.ElasticsearchException:
            print("could not delete es data form mode %s" % alias)
        try:
            # delete all our lexicons in sql
            for name in conf_mgr.get_lexiconlist(alias):
                db.deletebulk(lexicon=name)
        except Exception:
            print("could not delete sql data form mode %s" % alias)
    print("Successfully deleted all data")


def delete_mode(mode):
    # delete all indices
    es = conf_mgr.elastic(mode)
    try:
        # print('delete', '%s*' % mode)
        es.indices.delete("%s*" % mode)
    except esExceptions.ElasticsearchException:
        print("could not delete es data form mode %s" % mode)
    try:
        # delete all our lexicons in sql
        for name in conf_mgr.get_lexiconlist(mode):
            db.deletebulk(lexicon=name)
    except Exception:
        print("could not delete sql data form mode %s" % mode)
    print("Successfully deleted all data")




from .doc_converter import DocConverter
from .doc_converter import doc_to_es
from .doc_converter import doc_to_sql

from .autoupdates import auto_update  # noqa: F401
from .autoupdates import auto_update_child  # noqa: F401
from .autoupdates import auto_update_document  # noqa: F401
from .autoupdates import autoupdate_child  # noqa: F401




import json


def get_json(client, path, status=(200, 300)):
    print("Calling '{}' ...".format(path))
    response = client.get(path)
    assert status[0] <= response.status_code < status[1]
    return json.loads(response.data.decode())


def test_route_modes(client):
    r = get_json(client, "/modes")

    assert "default" in r
    assert "panacea" in r
    assert "karp" in r



import json

import pytest


def get_json(client, path):
    print("Calling '{}' ...".format(path))
    response = client.get(path)
    assert 200 <= response.status_code < 300
    return json.loads(response.data.decode())


@pytest.mark.parametrize(
    "q,mode,n_hits",
    [("sit", "panacea", 2), ("set", "panacea", 26), ],
)
def test_autocomplete(client_w_panacea, q, mode, n_hits):
    query = f"/autocomplete?q={q}&mode={mode}"
    result = get_json(client_w_panacea, query)

    print(f"result = {result}")
    assert result is not None
    assert "hits" in result
    assert "hits" in result["hits"]
    hits = result["hits"]["hits"]
    assert len(hits) == n_hits
    assert result["hits"]["total"] == n_hits



""" Class for exceptions generated by the backend, the converts, ElasticSearch
    or the auth server.
    Errors of type KarpException will be shown to the user
"""


class KarpException(Exception):
    """ The super class, copied from flask:
        http://flask.pocoo.org/docs/0.10/patterns/apierrors/
    """

    def __init__(self, message, debug_msg=None, status_code=None, user_msg=None, payload=None):
        super().__init__()
        self.message = message
        self.debug_msg = debug_msg or message
        self.user_msg = user_msg
        self.status_code = status_code or 400
        self.payload = payload

    def to_dict(self):
        rv = dict(self.payload or ())
        rv["message"] = self.message
        return rv

    def __str__(self):
        return "%s %s" % (self.status_code, self.message)


class KarpAuthenticationError(KarpException):
    """ Used for errors given by the auth server """

    def __init__(self, message, debug_msg=None, status_code=None, payload=None):
        super().__init__(
            "Authentication Exception: " + message,
            debug_msg=debug_msg,
            status_code=status_code or 401,
            payload=payload,
        )


class KarpElasticSearchError(KarpException):
    """ Used for errors given by the elastic search """

    def __init__(self, message, debug_msg=None, status_code=None, payload=None):
        KarpException.__init__(
            self, "Database Exception: " + message, debug_msg, status_code, payload
        )


class KarpDbError(KarpException):
    """ Used for errors given by the sql data base """

    def __init__(self, message, debug_msg=None, status_code=None, payload=None):
        KarpException.__init__(self, "SQL Error: %s" % message, debug_msg, status_code, payload)


class KarpParsingError(KarpException):
    """ Used for parsing errors, given during upload. """

    def __init__(self, message, debug_msg=None, status_code=None, payload=None):
        KarpException.__init__(
            self,
            "Parsing Error (no documents uploaded): %s" % message,
            debug_msg,
            status_code,
            payload,
        )


class KarpQueryError(KarpException):
    """ Used for errors given when trying to parse the query string. """

    def __init__(
        self, message, query=None, debug_msg=None, status_code=None, user_msg=None, payload=None,
    ):
        if query:
            msg = f"Query was '{query}'"
        else:
            msg = "No query."
        super().__init__(
            f"Query Error: {message}. {msg}",
            debug_msg=debug_msg,
            status_code=status_code,
            payload=payload,
            user_msg=user_msg,
        )


class KarpGeneralError(KarpException):
    """
    Used for unspecified errors. The actual error message will
    only be shown in the debug log. The user will just get a generic
    error message (see backend.py)
    """

    # TODO split into meaningful subclasses, some should not be shown to client
    # Note: all General Errors are invisible to client for now

    def __init__(
        self, message, user_msg=None, debug_msg=None, query=None, status_code=None, payload=None,
    ):
        """ message: is shown when exception is raised
            debug_msg: will only be shown in log (level=debug)
            user_msg: will be send to user
        """
        if not message:
            message = "Unknown error."
        if query:
            message = f"Error: {message}. Query was: {query}"
        else:
            message = f"Error: {message}"
        super().__init__(
            message,
            debug_msg=debug_msg,
            status_code=status_code,
            user_msg=user_msg,
            payload=payload,
        )



BYTE_TYPES = (bytes, bytearray)
TEXT_TYPES = (str,)


def uton(x):
    return x


def ntob(x):
    return x.encode("utf-8", "surrogateescape")


def ntou(x):
    return x


def bton(x):
    return x.decode("utf-8", "surrogateescape")


ESCAPE_TM = dict((k, ntou(repr(chr(k))[1:-1])) for k in range(32))
ESCAPE_TM[0] = u"\0"
ESCAPE_TM[7] = u"\a"
ESCAPE_TM[8] = u"\b"
ESCAPE_TM[11] = u"\v"
ESCAPE_TM[12] = u"\f"
ESCAPE_TM[ord("\\")] = u"\\\\"


def escape_control(s):
    if isinstance(s, TEXT_TYPES):
        return s.translate(ESCAPE_TM)
    else:
        return (
            s.decode("utf-8", "surrogateescape")
            .translate(ESCAPE_TM)
            .encode("utf-8", "surrogateescape")
        )


def unescape_control(s):
    if isinstance(s, TEXT_TYPES):
        return s.encode("latin1", "backslashreplace").decode("unicode_escape")
    else:
        return (
            s.decode("utf-8", "surrogateescape")
            .encode("latin1", "backslashreplace")
            .decode("unicode_escape")
            .encode("utf-8", "surrogateescape")
        )


# Replace with a less hacky function? originally from:
# https://stackoverflow.com/questions/9778550/which-is-the-correct-way-to-encode-escape-characters-in-python-2-without-killing
# def control_escape(s):
# """ Escapes control characters so that they can be parsed by a json parser.
# Eg. '\u0001' => '\\u0001'
# Note that u'\u0001'.encode('unicode_escape') will encode the string as
#'\\x01', which do not work for json. Hence the .replace('\\x', '\u00').
# """
# the set of characters migth need to be extended
# if type(s) is not str and type(s) is not str:
#    s = str(s)
# control_chars = [chr(c) for c in range(0x20)]
# return u''.join([c.encode('unicode_escape').replace('\\x', '\u00')
#                if c in control_chars else c for c in s])



__version__ = "5.24.11"

import pkg_resources
import logging
import logging.handlers
import os

import six

from .instance_info import get_instance_path

from flask import Flask, request

from karp5.config import Config, conf_mgr


__name = "karp5"


class RequestFormatter(logging.Formatter):
    def __init__(self, fmt=None, datefmt=None):
        logging.Formatter.__init__(self, fmt, datefmt)

    def format(self, record):
        record.req_url = request.url
        record.req_remote_addr = request.remote_addr
        record.req_method = request.method
        return logging.Formatter.format(self, record)


def create_app(config_class=Config):
    app = Flask(__name, instance_path=get_instance_path())
    app.config.from_object(config_class)

    conf_mgr.app_config = config_class

    conf_mgr.update_version(__version__)
    print("app.config = {}".format(app.config))

    # if app.config['ELASTICSEARCH_URL']:
    #     conf_mgr.override_elastic_url(app.config['ELASTICSEARCH_URL'])
    from karp5.context import auth

    if config_class.TESTING:
        auth.init("dummy")
    else:
        auth.init("std")

    from karp5.server.helper import flaskhelper

    flaskhelper.init_errorhandler(app)

    from karp5 import backend

    flaskhelper.register(app, backend.init)

    print("app.debug = {}".format(app.debug))
    print("app.testing = {}".format(app.testing))
    print("Setting up loggers")
    logger = logging.getLogger("karp5")
    logger.setLevel(app.config["LOG_LEVEL"])
    print("log_fmt = {}".format(app.config["LOG_FMT"]))
    print("log_datefmt = {}".format(app.config["LOG_DATEFMT"]))
    formatter = logging.Formatter(fmt=app.config["LOG_FMT"], datefmt=app.config["LOG_DATEFMT"])

    request_logger = logging.getLogger("karp5.requests")
    request_logger.setLevel(logging.INFO)
    request_formatter = RequestFormatter(
        fmt="[%(asctime)s] %(req_remote_addr)s %(req_method)s %(req_url)s %(message)s",
        datefmt=app.config["LOG_DATEFMT"],
    )
    if app.debug or app.config["LOG_TO_STDERR"]:
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(app.config["LOG_LEVEL"])
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)

        request_stream_handler = logging.StreamHandler()
        request_stream_handler.setFormatter(request_formatter)
        request_logger.addHandler(request_stream_handler)
    else:
        log_dir = app.config["LOG_DIR"]
        if not os.path.exists(log_dir):
            os.mkdir(log_dir)

        file_handler = logging.handlers.TimedRotatingFileHandler(
            os.path.join(log_dir, "karp5.log"), when="d", interval=1, backupCount=0
        )
        file_handler.setLevel(app.config["LOG_LEVEL"])
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

        request_file_handler = logging.handlers.TimedRotatingFileHandler(
            os.path.join(log_dir, "karp5.requests.log"), when="d", interval=1, backupCount=0,
        )
        request_file_handler.setFormatter(request_formatter)
        request_logger.addHandler(request_file_handler)

    @app.after_request
    def after_request(response):
        """ Logging after every request. """
        request_logger.info(response.status)
        return response

    return app


def get_version():
    return __version__


def get_name():
    return __name


def get_pkg_resource(filename):
    assert pkg_resources.resource_exists(__name__, filename)
    result = pkg_resources.resource_string(__name__, filename)
    if isinstance(result, six.binary_type):
        result = result.decode("utf-8")
    return result



from time import strftime

from elasticsearch import Elasticsearch
from elasticsearch import helpers as es_helpers
from elasticsearch import exceptions as esExceptions
import elasticsearch_dsl as es_dsl

import pytest

from karp5.config import mgr as conf_mgr
# from karp5.cli import upload_offline as upload


@pytest.mark.skip(reason="Only run separately.")
def test_scan():
    alias = "panacea"
    es = Elasticsearch(["http://localhost:9200"])
    target_index = "panacea_{}".format(strftime("%Y%m%d-%H%M%S"))
    data = conf_mgr.get_mapping(alias)
    try:
        ans = es.indices.create(index=target_index, body=data, request_timeout=30)
    except esExceptions.TransportError as e:
        raise Exception("Could not create index")

    print(ans)
    source_docs = es_helpers.scan(es, size=10000, index=alias, raise_on_error=True)

    def update_index(doc):
        doc["_index"] = target_index
        return doc

    updated_index = (update_index(doc) for doc in source_docs)
    success, failed, total = 0, 0, 0
    errors = []
    for ok, item in es_helpers.streaming_bulk(
        es, actions=updated_index, index=target_index
    ):
        if not ok:
            failed += 1
            errors.append(item)
        else:
            success += 1
            print("ok = {},item = {}".format(ok, item))
        total += 1
    # TODO when elasticsearch is updated to >=2.3: use es.reindex instead
    # ans = es_helpers.reindex(es, source_index, target_index)
    assert success == total


@pytest.mark.skip(reason="Doesn't work")
def test_dsl_scan():
    alias = "panacea"
    es = Elasticsearch(["http://localhost:9200"])
    s = es_dsl.Search(using=es, index=alias)
    target_index = "panacea_{}".format(strftime("%Y%m%d-%H%M%S"))
    data = conf_mgr.get_mapping(alias)
    try:
        ans = es.indices.create(index=target_index, body=data, request_timeout=30)
    except esExceptions.TransportError as e:
        raise Exception("Could not create index")

    print(ans)
    source_docs = s.scan()

    # for n, doc in enumerate(source_docs, 1):
    #     print(doc)
    #     assert '_source' in doc

    # assert n == 0
    def update_index(doc):
        doc["_index"] = target_index
        return doc

    updated_index = (update_index(doc) for doc in source_docs)
    success, failed, total = 0, 0, 0
    errors = []
    for ok, item in es_helpers.streaming_bulk(
        es, actions=source_docs, index=target_index
    ):
        if not ok:
            failed += 1
            errors.append(item)
        else:
            success += 1
            print("ok = {},item = {}".format(ok, item))
        total += 1
    # TODO when elasticsearch is updated to >=2.3: use es.reindex instead
    # ans = es_helpers.reindex(es, source_index, target_index)
    assert success != total



import pytest


from karp5.tests.util import get_json


@pytest.mark.parametrize("command", ["query", "querycount", "minientry",])
@pytest.mark.parametrize(
    "q,mode,n_hits",
    [
        ("simple||sit", "panacea", 2),
        ("simple||sit", "karp", 2),
        ("extended||and|pos|exists", "panacea", 6609),
        ("extended||and|pos|equals|Vb", "panacea", 1677),
        ("extended||and|pos|missing", "panacea", 0),
        ("extended||and|baseform|startswith|ang", "panacea", 22),
        ("extended||and|baseform|startswith|ang|be", "panacea", 422),
        ("extended||and|pos|equals|Vb||and|baseform|startswith|ab", "panacea", 65),
        ("extended||and|baseform_en|regexp|s.*t", "panacea", 313),
        ("extended||and|pos|equals|Vb||not|baseform|startswith|ab", "panacea", 1612),
    ],
)
def test_search(client_w_panacea, command, q, mode, n_hits):
    query = f"/{command}?q={q}&mode={mode}"
    result = get_json(client_w_panacea, query)

    if command == "querycount":
        assert "distribution" in result
        if n_hits == 0:
            assert len(result["distribution"]) == 0
        else:
            assert result["distribution"][0]["doc_count"] == n_hits
            assert len(result["distribution"]) == 1
            assert len(result["distribution"][0]["lexiconName"]["buckets"]) == 1

        assert "query" in result
        assert "hits" in result["query"]
        assert "total" in result["query"]["hits"]
        assert result["query"]["hits"]["total"] == n_hits
    else:
        assert "hits" in result
        assert "total" in result["hits"]
        assert result["hits"]["total"] == n_hits



from karp5.server.translator import parser

from karp5.tests.util import assert_es_search


def test_search_get_context(app):
    exps = [{"match_phrase": {"lexiconName": "panacea"}}]
    with app.test_request_context("/getcontext/panacea"):
        result = parser.search(exps, [], [], usefilter=True, constant_score=False)
    expected = {"query": {"bool": {"must": [{"match_phrase": {"lexiconName": "panacea"}}]}}}
    expected = {"bool": {"must": [{"match_phrase": {"lexiconName": "panacea"}}]}}
    assert_es_search(result, expected)
    assert result == expected


def test_search_get_pre_post(app):
    exps = [
        {"match_phrase": {"lexiconName": "panacea"}},
        {"range": {"lemma_german.raw": {"gte": "Abbau"}}},
    ]
    with app.test_request_context("/getcontext/panacea"):
        result = parser.search(exps, [], [], usefilter=False, constant_score=True)
    expected = {
        "query": {
            "bool": {
                "filter": [
                    {"match_phrase": {"lexiconName": "panacea"}},
                    {"range": {"lemma_german.raw": {"gte": "Abbau"}}},
                ]
            }
        }
    }
    expected = {
        "query": {
            "bool": {
                "must": [
                    {"match_phrase": {"lexiconName": "panacea"}},
                    {"range": {"lemma_german.raw": {"gte": "Abbau"}}},
                ]
            }
        }
    }
    assert_es_search(result, expected)
    assert result == expected


def test_search_autocomplete_q(app):
    exp = {
        "bool": {
            "should": [
                {"term": {"lemma_german": {"boost": "500", "value": "sig"}}},
                {"match_phrase": {"lemma_german": "sig"}},
                {"match_phrase": {"english.lemma_english": "sig"}},
            ]
        }
    }
    fields = {"exists": {"field": "lemma_german"}}
    resource = {
        "bool": {
            "should": [
                {"term": {"lexiconName": "bar"}},
                {"term": {"lexiconName": "foo"}},
                {"term": {"lexiconName": "large_lex"}},
                {"term": {"lexiconName": "panacea"}},
                {"term": {"lexiconName": "panacea_links"}},
            ]
        }
    }
    exps = [exp, fields, resource]
    filters = []
    fields = ""
    with app.test_request_context("/autocomplete?q=sig"):
        result = parser.search(exps, filters, fields, usefilter=True)
    expected = {
        "query": {
            "bool": {
                "filter": [
                    {
                        "bool": {
                            "should": [
                                {"term": {"lemma_german": {"boost": "500", "value": "sig"}}},
                                {"match_phrase": {"lemma_german": "sig"}},
                                {"match_phrase": {"english.lemma_english": "sig"}},
                            ]
                        }
                    },
                    {"exists": {"field": "lemma_german"}},
                    {
                        "bool": {
                            "should": [
                                {"term": {"lexiconName": "bar"}},
                                {"term": {"lexiconName": "foo"}},
                                {"term": {"lexiconName": "large_lex"}},
                                {"term": {"lexiconName": "panacea"}},
                                {"term": {"lexiconName": "panacea_links"}},
                            ]
                        }
                    },
                ]
            }
        }
    }
    expected = {
        "query": {
            "constant_score": {
                "filter": {
                    "bool": {
                        "must": [
                            {
                                "bool": {
                                    "should": [
                                        {
                                            "term": {
                                                "lemma_german": {"boost": "500", "value": "sig"}
                                            }
                                        },
                                        {"match_phrase": {"lemma_german": "sig"}},
                                        {"match_phrase": {"english.lemma_english": "sig"}},
                                    ]
                                }
                            },
                            {"exists": {"field": "lemma_german"}},
                            {
                                "bool": {
                                    "should": [
                                        {"term": {"lexiconName": "bar"}},
                                        {"term": {"lexiconName": "foo"}},
                                        {"term": {"lexiconName": "large_lex"}},
                                        {"term": {"lexiconName": "panacea"}},
                                        {"term": {"lexiconName": "panacea_links"}},
                                    ]
                                }
                            },
                        ]
                    }
                }
            }
        }
    }
    assert_es_search(result, expected)
    assert result == expected




from datetime import timedelta
import logging
import six
from elasticsearch import ConnectionError
from flask import make_response, request, current_app
from functools import update_wrapper

from karp5 import errors
from karp5.config import mgr as conf_mgr
import karp5.server.update as update
from karp5.dbhandler import emailsender as email


_logger = logging.getLogger("karp5")


# Decorator for allowing cross site http request etc.
# http://flask.pocoo.org/snippets/56/
def crossdomain(
    origin=None,
    methods=None,
    headers=None,
    max_age=21600,
    attach_to_all=True,
    automatic_options=True,
):
    if methods is not None:
        methods = ", ".join(sorted(x.upper() for x in methods))
    if headers is None:
        # Set standard headers here
        # TODO figure out which ones that are meaningful (when)
        headers = ["Content-Type", "Authorization"]
    if headers is not None and not isinstance(headers, six.string_types):
        headers = ", ".join(x.upper() for x in headers)
    if not isinstance(origin, six.string_types):
        origin = ", ".join(origin)
    if isinstance(max_age, timedelta):
        max_age = max_age.total_seconds()

    def get_methods():
        if methods is not None:
            return methods

        options_resp = current_app.make_default_options_response()
        return options_resp.headers["allow"]

    def decorator(f):
        def wrapped_function(*args, **kwargs):
            if automatic_options and request.method == "OPTIONS":
                resp = current_app.make_default_options_response()
            else:
                resp = make_response(f(*args, **kwargs))
            if not attach_to_all and request.method != "OPTIONS":
                return resp

            h = resp.headers

            h["Access-Control-Allow-Origin"] = origin
            h["Access-Control-Allow-Methods"] = get_methods()
            h["Access-Control-Max-Age"] = str(max_age)
            if headers is not None:
                h["Access-Control-Allow-Headers"] = headers
            return resp

        f.provide_automatic_options = False
        f.required_methods = ["OPTIONS"]
        return update_wrapper(wrapped_function, f)

    return decorator


def register(app, initiator):
    urls = []

    def route(url="", methods=None, crossdomain=True, name=None):
        """ Decorator function @route
            Adds the function to a list of urls, which should later be processed
            by Flask (see flaskhelper.py)
            If 'name' is not given, the url will be the name of the function.
            If 'url', it will appended to the function name, and may contain variable
            parts of the path
            'methods' is a list of allowed methods, defaults to 'GET'
            Example:
            @route()
            def mypage():
               return render_page("hello")
            ==> /mypage

            @route('<pagename>')
            def mypage(pagename=''):
               return render_page("welcome to"+pagename)
            ==> /mypage/any_page_name
        """

        def f(func):
            if name is not None:
                urlname = name
            elif url:
                urlname = "/%s/%s" % (func.__name__, url)
            else:
                urlname = "/%s" % func.__name__
            _logger.debug("add url\n\n")
            urls.append((urlname, func, methods, crossdomain))
            return func

        return f

    initiator(route)
    for url, func, methods, cross in urls:
        if cross:
            # add the crossdomain decorator to the view function
            func = crossdomain(origin="*", methods=methods)(func)
        app.add_url_rule(url, endpoint=url, view_func=func, methods=methods)


# TODO test if error handling works. If not: move the decorator to top level
# Error handling, show all KarpExceptions to the client
def init_errorhandler(app):
    @app.errorhandler(Exception)
    @crossdomain(origin="*")
    def handle_invalid_usage(error):
        try:
            request.get_data()
            data = request.data
            data = data.decode("utf8")
            auth = request.authorization
            e_type = "Predicted" if isinstance(error, errors.KarpException) else "Unpredicted"

            _logger.debug("Error on url %s" % request.full_path)
            user = "unknown"
            if auth:
                user = auth.username
            # s = '%s: %s  User: %s\n\t%s: %s\n\t%s\n' \
            #     % (datetime.datetime.now(), request.path, user, e_type,
            #        str(error), data)
            s = "%s  User: %s\n%s: %s\n%s\n" % (request.full_path, user, e_type, str(error), data,)
            _logger.exception(s)

            if isinstance(error, ConnectionError):
                _logger.debug(update.handle_update_error(error, data, user, ""))

            if isinstance(error, errors.KarpException):
                # Log full error message if available
                if error.debug_msg:
                    _logger.debug(error.debug_msg)

                status_code = error.status_code
                if error.user_msg:
                    return str(error.user_msg), status_code
                else:
                    return error.message, status_code

            else:
                _logger.exception(error)
                return "Oops, something went wrong\n", 500

        except Exception:
            # In case of write conflicts etc, print to anoter file
            # and send email to admin
            import time
            import traceback

            logdir = conf_mgr.app_config.LOG_DIR
            trace = traceback.format_exc()
            date = time.strftime("%Y-%m-%d %H:%M:%S")
            msg = "Cannot print log file: %s, %s" % (date, trace)
            title = "Karp urgent logging error"
            if conf_mgr.app_config.ADMIN_EMAILS:

                email.send_notification(conf_mgr.app_config.ADMIN_EMAILS, title, msg)
            # open(logdir + "KARPERR" + time.strftime("%Y%m%d"), "a").write(msg)
            _logger.error(msg)
            return "Oops, something went wrong\n", 500



# -*- coding=UTF-8 -*-
"""
Connect to the sql data base and interact with it.
Emails the admins (config/config.json) if an error occurs.
"""

import datetime
import json
import logging
import traceback
from typing import Iterable, List, Tuple, Union, Optional

import sqlalchemy as sql
from sqlalchemy.ext.compiler import compiles

import karp5
from karp5.config import mgr as conf_mgr
from karp5.errors import KarpGeneralError


_logger = logging.getLogger("karp5")


@compiles(sql.VARCHAR, "mysql")
@compiles(sql.String, "mysql")
def compile_varchar(element, compiler, **kw):
    """ Forces mysql to use case sensitiveness for strings types """
    return "VARCHAR(%s) COLLATE utf8_bin" % (element.length)


STATUS_CHANGE = sql.types.Enum("added", "changed", "removed", "imported")
STATUS_SUGG = sql.types.Enum("waiting", "accepted", "rejected", "accepted_modified")


def get_engine(lexicon, mode=None, suggestion=False, echo=True):
    if mode:
        dburl = conf_mgr.get_mode_sql(mode)
    else:
        dburl = conf_mgr.get_lexicon_sql(lexicon)
    print("dburl = {}".format(dburl))
    if not dburl:
        raise SQLNull("%s/%s" % (lexicon, mode))

    metadata = sql.MetaData()
    if not suggestion:
        db_entry = create_table(metadata)
    else:
        db_entry = create_suggestion_table(metadata)

    engine = sql.create_engine(dburl, encoding="utf-8", echo=echo)
    metadata.create_all(engine)
    return engine, db_entry


def create_table(metadata):
    db_entry = sql.Table(
        "karpentry",
        metadata,
        sql.Column("id", sql.String(50), index=True),
        sql.Column("date", sql.types.DateTime, index=True),
        sql.Column("user", sql.String(320), index=True),
        # Text(2**24-1) corresponds to MediumText in MySQL
        # avoid using the type MediumText (specific to MySQL)
        sql.Column("source", sql.types.Text(2 ** 24 - 1)),
        sql.Column("msg", sql.String(160)),
        sql.Column("lexicon", sql.String(50), index=True),
        sql.Column("status", STATUS_CHANGE),
        sql.Column("version", sql.Integer),
    )
    sql.Index("historyindex", db_entry.c.lexicon, db_entry.c.status, db_entry.c.date)
    return db_entry


def create_suggestion_table(metadata):
    db_entry = sql.Table(
        "karpsuggestions",
        metadata,
        sql.Column("id", sql.String(50), index=True),
        sql.Column("date", sql.types.DateTime, index=True),
        sql.Column("user", sql.String(320), index=True),
        # Text(2**24-1) corresponds to MediumText in MySQL
        # avoid using the type MediumText (specific to MySQL)
        sql.Column("source", sql.types.Text(2 ** 24 - 1)),
        sql.Column("msg", sql.String(160)),
        sql.Column("lexicon", sql.String(50), index=True),
        # For suggestions
        sql.Column("status", STATUS_SUGG, index=True),
        sql.Column("origid", sql.String(22), index=True),
        # Remember which version this is a copy of
        sql.Column("version", sql.Integer),
        sql.Column("acceptmsg", sql.String(160)),
    )
    return db_entry


def update_test(_id, lexicon, doc, user, msg):
    try:
        engine, db_entry = get_engine(lexicon)
        ins = db_entry.insert().values(
            id=_id, date=datetime.datetime.now(), user=user, msg=msg, source=doc
        )
        conn = engine.connect()
        conn.execute(ins)
        return 1, ""
    except SQLNull as e:
        return 0, (e.message)
    except Exception as e:
        return 0, handle_error(e, user, msg, doc)


def update(
    _id,
    doc,
    user,
    msg,
    lexicon,
    version=0,
    status="waiting",
    engine=None,
    db_entry=None,
    suggestion_id="",
    date="",
):
    """ Puts an update in the database. If several updates are to be done, an
        engine should be created beforehand in order to avoid errors due to too
        many connections (1040). If no engine is provided, a new one is created
        If suggestion_id is set, this is added to the suggestion table, with
        _id used as orginid.
    """
    try:
        if engine is None:
            engine, db_entry = get_engine(lexicon, suggestion=bool(suggestion_id))

        try:
            version = int(version)
        except ValueError as e:
            version = -1
        if suggestion_id:
            # If the suggestion_id is a bool, set it to be empty
            # A string here means that the suggestion is a modification,
            # while True means that it is a suggested addition.
            # No suggestion_id at all (empty string or False) would mean that
            # it is not a suggestion but a real update.
            if isinstance(suggestion_id, bool):
                suggestion_id = ""
            ins = db_entry.insert().values(
                id=_id,
                origid=suggestion_id,
                date=date or datetime.datetime.now(),
                user=user,
                source=doc,
                version=version,
                msg=msg,
                acceptmsg="",
                status="waiting",
                lexicon=lexicon,
            )
        else:
            ins = db_entry.insert().values(
                id=_id,
                lexicon=lexicon,
                date=date or datetime.datetime.now(),
                user=user,
                msg=msg,
                source=doc,
                status=status,
                version=version,
            )
        conn = engine.connect()
        conn.execute(ins)
        conn.close()
        return 1, ""
    except SQLNull as e:
        return 0, e.message
    except Exception as e:
        return 0, handle_error(e, user, msg, doc)


def update_bulk(
    lexicon: str, bulk: Iterable[Tuple[str, str, str, str, str, str]]
) -> Tuple[int, str]:
    user = "admin"
    try:
        engine, db_entry = get_engine(lexicon, echo=False)
        gen_bulk = []
        for (_id, data, user, msg, lex, status) in bulk:
            user = user
            gen_bulk.append(
                {
                    "id": _id,
                    "date": datetime.datetime.now(),
                    "user": user,
                    "source": data,
                    "msg": msg,
                    "lexicon": lex,
                    "status": status,
                }
            )

        # if outputfile:
        #     with open(outputfile,'w') as f:
        #        for ins in gen_bulk:
        #            s = db_entry.insert(ins).compile(engine,
        #                             compile_kwargs={"literal_binds" : True})
        #            f.write(str(s))
        # else:
        engine.execute(db_entry.insert(), gen_bulk)
        return len(gen_bulk), ""

    except SQLNull as e:
        return 0, e.message
    except Exception as e:
        return 0, handle_error(e, user, "bulk update: %s" % e, None)


def dbselect(
    lexicon,
    user="",
    _id="",
    from_date="",
    to_date="",
    exact_date="",
    status=None,
    max_hits=10,
    engine=None,
    db_entry=None,
    suggestion=False,
    mode="",
):
    # does not accept a list of lexicons anymore
    if status is None:
        status = []
    try:
        if engine is None or db_entry is None:
            engine, db_entry = get_engine(lexicon, mode=mode, suggestion=suggestion)

        conn = engine.connect()
        operands = []
        if user:
            operands.append(db_entry.c.user == user)
        if _id:
            operands.append(db_entry.c.id == _id)
        if from_date:
            operands.append(db_entry.c.date >= from_date)
        if to_date:
            operands.append(db_entry.c.date <= to_date)
        if exact_date:
            operands.append(db_entry.c.date == exact_date)
        if lexicon:
            operands.append(db_entry.c.lexicon == lexicon)
        add_list_operands([(status, db_entry.c.status)], operands)
        selects = sql.select([db_entry]).where(sql.and_(*operands))
        if max_hits > 0:
            selects = selects.limit(max_hits)  # only get the first hits
        selects = selects.order_by(db_entry.c.date.desc())  # sort by date
        _logger.info("selects = %s", selects)
        res = []
        for entry in conn.execute(selects):
            # transform the date into a string now to enforce isoformat
            version = 8 if suggestion else 7
            obj = {
                "id": entry[0],
                "date": str(entry[1]),
                "user": entry[2],
                "doc": json.loads(entry[3]),
                "lexicon": entry[5],
                "message": entry[4],
                "status": entry[6],
                "version": entry[version],
            }
            if suggestion:
                obj["acceptmessage"] = entry[9]
                obj["origid"] = entry[7]
            res.append(obj)
        conn.close()
        return res

    except SQLNull as e:
        _logger.warning("Attempt to search in SQL, no db available: %s", e.message)
        return []


def dbselect_gen(
    lexicon,
    user="",
    _id="",
    from_date="",
    to_date="",
    exact_date="",
    status=None,
    max_hits=10,
    engine=None,
    db_entry=None,
    suggestion=False,
    mode="",
):
    # does not accept a list of lexicons anymore
    if status is None:
        status = []
    try:
        if engine is None or db_entry is None:
            engine, db_entry = get_engine(lexicon, mode=mode, suggestion=suggestion)

        conn = engine.connect()
        operands = []
        if user:
            operands.append(db_entry.c.user == user)
        if _id:
            operands.append(db_entry.c.id == _id)
        if from_date:
            operands.append(db_entry.c.date >= from_date)
        if to_date:
            operands.append(db_entry.c.date <= to_date)
        if exact_date:
            operands.append(db_entry.c.date == exact_date)
        if lexicon:
            operands.append(db_entry.c.lexicon == lexicon)
        add_list_operands([(status, db_entry.c.status)], operands)
        selects = sql.select([db_entry]).where(sql.and_(*operands))
        if max_hits > 0:
            selects = selects.limit(max_hits)  # only get the first hits
        selects = selects.order_by(db_entry.c.date.desc())  # sort by date
        _logger.info("selects = %s", selects)
        for entry in conn.execute(selects):
            # transform the date into a string now to enforce isoformat
            version = 8 if suggestion else 7
            obj = {
                "id": entry[0],
                "date": str(entry[1]),
                "user": entry[2],
                "doc": json.loads(entry[3]),
                "lexicon": entry[5],
                "message": entry[4],
                "status": entry[6],
                "version": entry[version],
            }
            if suggestion:
                obj["acceptmessage"] = entry[9]
                obj["origid"] = entry[7]
            yield obj
        conn.close()
        # return res

    except SQLNull as e:
        _logger.warning("Attempt to search in SQL, no db available: %s", e.message)
        return


def add_list_operands(to_add, operands):
    for vals, row_val in to_add:
        disjunct_operands = []
        if isinstance(vals, str):
            vals = [vals]
        for val in vals:
            disjunct_operands.append(row_val == val)

        operands.append(sql.or_(*disjunct_operands))


def get_entries_to_keep_gen(lexicon, *, to_date=None):
    engine, db_entry = get_engine(lexicon, echo=False)

    _logger.debug("exporting entries from %s ", lexicon)

    dbselect_kwargs = {"engine": engine, "db_entry": db_entry, "max_hits": -1}
    if to_date is not None:
        dbselect_kwargs["to_date"] = to_date

    processed_ids = set()
    for entry in dbselect_gen(lexicon, **dbselect_kwargs):
        if entry["id"] in processed_ids:
            # print(f"skipping entry = {entry}")
            continue
        elif entry["status"] == "removed":
            # print(f"skipping entry = {entry}")
            processed_ids.add(entry["id"])
            continue
        else:
            processed_ids.add(entry["id"])
            yield entry


def modifysuggestion(_id, lexicon, msg="", status="", origid="", engine=None, db_entry=None):
    try:
        if engine is None or db_entry is None:
            engine, db_entry = get_engine(lexicon, suggestion=True)
        if isinstance(msg, str):
            msg = msg.encode("utf-8")
        conn = engine.connect()
        operands = []
        if status:
            operands.append(db_entry.c.status == status)
        if msg:
            operands.append(db_entry.c.msg == msg)
        if origid:
            operands.append(db_entry.c.origid == origid)
        update_cmd = (
            db_entry.update()
            .where(db_entry.c.id == _id)
            .values({"status": status, "acceptmsg": msg})
        )
        conn.execute(update_cmd)
        conn.close()
        return 1, ""

    except Exception as e:
        return 0, handle_error(e, "--modification--", msg, "--modified--")


def handle_error(e, user, msg, doc):
    _logger.exception(e)
    mail_sent = "No warnings sent by email"
    if conf_mgr.app_config.ADMIN_EMAILS:
        import karp5.dbhandler.emailsender as sender

        # report = "User: %s, msg: %s. \nDoc:\n%s" % (user, msg, doc)
        msg = f"""Karp5-backend failure
        Time: {datetime.datetime.now()}
        Version: {karp5.get_version()}
        Error: {e}
        User: {user}
        Message: {msg}
        Document: {doc}
        Traceback: {traceback.format_exc()}"""
        sender.send_notification(conf_mgr.app_config.ADMIN_EMAILS, "Karp5-backend failure", msg)
        mail_sent = "Warning sent to %s" % ", ".join(conf_mgr.app_config.ADMIN_EMAILS)
    return "%s. %s" % (str(e), mail_sent)


def delete(lexicon, _id):
    engine, dbtable = get_engine(lexicon)
    conn = engine.connect()
    conn.execute(dbtable.delete().where(dbtable.c.id == _id))
    conn.close()

    return []


def deletebulk(lexicon: Optional[str] = None, user: Optional[str] = None):
    engine, dbtable = get_engine(lexicon)
    conn = engine.connect()
    operands = []
    if user and lexicon:
        operands = []
        operands.append(dbtable.c.user == user)
        operands.append(dbtable.c.lexicon == lexicon)
        choice = sql.and_(*operands)
    elif user:
        choice = dbtable.c.user == user
    elif lexicon:
        choice = dbtable.c.lexicon == lexicon
    else:
        raise DBUsageError("Must give either 'lexicon' or 'user', or both.")

    conn.execute(dbtable.delete().where(choice))
    conn.close()


def get_entries_to_keep(lexicons: Union[str, List[str]], *, to_date=None, exclude_states=None):
    """Retrieve entries from one or several lexicons to keep.

    Arguments:
        lexicons {Union[str, List[str]]} -- lexicon or lexicons to export

    Keyword Arguments:
        to_date {[type]} -- Optional date to limit the search (default: {None})

    Returns:
        Iterable -- [descrip
    """
    to_keep = {}
    if isinstance(lexicons, str):
        lexicons = [lexicons]

    for lex in lexicons:
        engine, db_entry = get_engine(lex, echo=False)
        dbselect_kwargs = {"engine": engine, "db_entry": db_entry, "max_hits": -1}
        if to_date:
            dbselect_kwargs["to_date"] = to_date
        for entry in dbselect(lex, **dbselect_kwargs):
            _id = entry["id"]
            if _id:  # don't add things without id, they are errors
                if _id in to_keep:
                    last = to_keep[_id]["date"]
                    if last < entry["date"]:
                        _logger.debug("|get_entries_to_keep_from_sql| Update entry.")
                        to_keep[_id] = entry
                else:
                    to_keep[_id] = entry
            else:
                _logger.warning("|sql no id| Found entry without id:")
                _logger.warning("|sql no id| %s", entry)
    # _logger.debug("to_keep = %s", to_keep)
    if exclude_states:
        if isinstance(exclude_states, str):
            exclude_states = [exclude_states]
        gen_out = ((i, v) for i, v in to_keep.items() if v["status"] not in exclude_states)
    else:
        gen_out = to_keep.items()
    return gen_out


class SQLNull(KarpGeneralError):
    """ Tells that there is no SQL instance """

    def __init__(self, lex):
        super().__init__(f"No SQL db available for {lex}")

    def __str__(self):
        return f"{self.message}"


class DBUsageError(KarpGeneralError):
    """ Error in usage of the dbhandler module. """

    def __init__(self, msg):
        super().__init__(msg)

    def __str__(self):
        return f"{self.message}"



from unittest import mock

from hypothesis import assume, given, settings, strategies as st

from karp5.tests.util import get_json


@settings(deadline=None)
@given(start=st.sampled_from([0, 1, 40, 200, 264]), size=st.just(25))
# @given(start=st.integers(min_value=0), size=st.integers(min_value=0))
def test_pagination_panacea(client_w_panacea, start, size):
    command = "query"
    q = "extended||and|baseform|regexp|.*"
    mode = "panacea"
    user_is_authenticated = True
    query = f"/{command}?q={q}&mode={mode}&start={start}&size={size}"
    with mock.patch(
        "karp5.context.auth.validate_user", return_value=(user_is_authenticated, [mode])
    ):
        result = get_json(client_w_panacea, query)

    assert result is not None
    assert "hits" in result
    assert "hits" in result["hits"]
    if start > 6609:
        calculated_size = 0
    else:
        size_start = start + size
        if size_start < 6609:
            calculated_size = size
        else:
            calculated_size = min(size, 6609) - start

    # calculated_size = max(min(size + start, 6609) - start - 1, 0)
    assert len(result["hits"]["hits"]) == calculated_size


@settings(deadline=None)
@given(start=st.sampled_from([0, 1, 40, 200, 1000, 1500]), size=st.just(25))
# @given(start=st.integers(min_value=0), size=st.integers(min_value=0))
def test_pagination_large_lex(client_w_large_lex, start, size):
    command = "query"
    q = "extended||and|foo|regexp|.*"
    mode = "large_lex"
    query = f"/{command}?q={q}&mode={mode}&start={start}&size={size}"

    result = get_json(client_w_large_lex, query)

    assert result is not None
    assert "hits" in result
    assert "hits" in result["hits"]
    if start > 6609:
        calculated_size = 0
    else:
        size_start = start + size
        if size_start < 6609:
            calculated_size = size
        else:
            calculated_size = min(size, 6609) - start

    # calculated_size = max(min(size + start, 6609) - start - 1, 0)
    assert len(result["hits"]["hits"]) == calculated_size



from karp5.domain.models.user import User


def test_empty_user():
    user = User()

    assert not user.is_authenticated
    assert user.is_not_authenticated()
    assert user.lexicon_permissions == {}
    assert user.name == ""



import attr
from typing import Dict


@attr.s(auto_attribs=True)
class User:
    name: str = ""
    is_authenticated: bool = False
    lexicon_permissions: Dict[str, Dict[str, bool]] = attr.Factory(dict)

    def is_not_authenticated(self) -> bool:
        return not self.is_authenticated



import pytest

import elasticsearch_dsl as es_dsl

from karp5.domain.services import search_service
from karp5.config import conf_mgr


@pytest.mark.parametrize("from_,size", [
    (0, 25),
    (1000, 25),
    (10000, 25),
    (0, None),
    (15000, 15875)
])
def test_large_lex(app_w_large_lex, from_, size):
    mode = "large_lex"
    total_num_entries = 20000
    es_search = es_dsl.Search(using=conf_mgr.elastic(mode), index=mode)
    result = search_service.execute_query(es_search, from_=from_, size=size)
    if size is None:
        expected_len_hits = total_num_entries - from_
    else:
        expected_len_hits = size if (from_ + size < total_num_entries) else total_num_entries - from_

    assert len(result["hits"]["hits"]) == expected_len_hits
    assert result["hits"]["total"] == total_num_entries
    for hit in result["hits"]["hits"]:
        assert "_source" in hit



from unittest import mock

import pytest

from karp5.server.translator import parser
from karp5.config import conf_mgr

from karp5.tests.util import assert_es_search  # pytype: disable=import-error


def test_make_settings_empty_call(app):
    settings = parser.make_settings(None, {}, user_is_authorized=False)

    assert settings["allowed"] is None
    assert settings["mode"] == conf_mgr.app_config.STANDARDMODE
    assert not settings["user_is_authorized"]


def test_make_settings_append_field():
    settings = parser.make_settings(None, {"extra": "added"}, user_is_authorized=True)

    assert settings["allowed"] is None
    assert settings["mode"] == conf_mgr.app_config.STANDARDMODE
    assert "extra" in settings
    assert settings["extra"] == "added"


def test_make_settings_change_mode():
    settings = parser.make_settings(None, {"mode": "new"}, user_is_authorized=True)

    assert settings["allowed"] is None
    assert settings["mode"] == "new"


@pytest.mark.parametrize("authorized", [False, True])
def test_make_settings_set_user_is_authorized(authorized):
    settings = parser.make_settings(None, {}, user_is_authorized=authorized)

    assert settings["user_is_authorized"] == authorized


def test_parse_extra_empty_call(app):
    with app.test_request_context("/query?q=simple||hej"):
        settings = {}
        with pytest.raises(parser.errors.AuthenticationError):
            parser.parse_extra(settings)


def test_parse_extra_unsupported_request_args(app):
    with app.test_request_context("/query?q=simple||hej&invalid=really"):
        settings = {}
        with pytest.raises(parser.errors.QueryError):
            parser.parse_extra(settings)


def test_parse_extra_minimum_successful(app):
    with app.test_request_context("/query?q=simple||hej"):
        settings = {"mode": "karp", "allowed": ["any"]}  # must exist in conf_mgr
        p_extra = parser.parse_extra(settings)

        assert p_extra == {"term": {"lexiconName": "any"}}


def test_construct_exp_empty_call():
    assert parser.construct_exp(None) == {}


@pytest.mark.parametrize("exps", ["a", "b"])
@pytest.mark.parametrize("querytype", ["filter", "must", "q"])
@pytest.mark.parametrize("constant_score", [True, False])
def test_construct_exp_string(exps, querytype, constant_score):
    result = parser.construct_exp(exps, querytype, constant_score)
    expected = {querytype: {"constant_score": exps} if constant_score else exps}
    assert result == expected


@pytest.mark.parametrize("exps", [["a", "b"]])
@pytest.mark.parametrize("querytype", ["filter", "must", "q"])
@pytest.mark.parametrize("constant_score", [True, False])
def test_construct_exp_list(exps, querytype, constant_score):
    result = parser.construct_exp(exps, querytype, constant_score)
    expected = {"must": exps}
    if querytype != "must":
        expected = {querytype: {"bool": expected}}
    assert result == expected


def test_parse_empty_call(app):
    with app.test_request_context("/query?q=a"):
        with pytest.raises(parser.errors.AuthenticationError):
            parser.parse({})


def test_parse_no_q(app):
    with app.test_request_context("/query"):
        settings = {"mode": "karp", "allowed": ["any"]}
        with pytest.raises(parser.errors.QueryError):
            parser.parse(settings)


def test_parse_q_simple_mode_karp(app):
    text = "hej"
    lexicon = "any"
    with app.test_request_context(f"/query?q=simple||{text}"):
        settings = {"mode": "karp", "allowed": [lexicon]}
        result = parser.parse(settings)
        expected = {
            "query": {
                "bool": {
                    "must": [
                        {
                            "bool": {
                                "should": [
                                    {"match": {"_all": {"operator": "and", "query": text}}},
                                    {"match": {"lemma_german": {"boost": 200, "query": text}}},
                                    {
                                        "match": {
                                            "english.lemma_english": {"boost": 100, "query": text}
                                        }
                                    },
                                ]
                            }
                        },
                        {"term": {"lexiconName": "any"}},
                    ]
                }
            }
        }
        assert_es_search(result, expected)
        assert result == expected


@pytest.mark.parametrize("user_is_authorized", [False, True])
def test_parse_q_simple_mode_foo(app, user_is_authorized):
    """Used in karp5.server.searching.requestquery when User is not authorized"""
    text = "hej"
    lexicon = "foo"
    with app.test_request_context(f"/query?q=simple||{text}"):
        settings = {"mode": lexicon, "allowed": [lexicon], "user_is_authorized": user_is_authorized}
        result = parser.parse(settings)
        expected = {
            "query": {
                "bool": {
                    "must": [
                        {
                            "bool": {
                                "should": [
                                    {"match": {"_all": {"operator": "and", "query": text}}},
                                    {"match": {"foo": {"boost": 100, "query": text}}},
                                ]
                            }
                        },
                        {"term": {"lexiconName": lexicon}},
                    ],
                }
            }
        }
        if not user_is_authorized:
            expected["query"]["bool"]["filter"] = {"term": {"status": "ok"}}
        assert_es_search(result, expected)
        assert result == expected


@pytest.mark.parametrize("user_is_authorized", [False, True])
def test_parse_q_extended_mode_foo(app, user_is_authorized):
    """Used in karp5.server.searching.requestquery when User is not authorized"""
    text = "hej"
    lexicon = "foo"
    settings = {"mode": lexicon, "allowed": [lexicon], "user_is_authorized": user_is_authorized}
    with app.test_request_context(f"/query?q=extended||and|foo|equals|{text}"):
        result = parser.parse(settings)

    expected_filter_must = [
        {"term": {"lexiconName": lexicon}},
        {"match_phrase": {"foo": text}},
    ]
    if not user_is_authorized:
        expected_filter_must.append({"term": {"status": "ok"}})
    expected = {"query": {"constant_score": {"filter": {"bool": {"must": expected_filter_must}}}}}
    assert_es_search(result, expected)
    assert result == expected


def test_search_query_with_score(app):
    lexicon = "panacea"
    text = "test"
    settings = {"mode": lexicon, "allowed": [lexicon], "sort": ["_score"]}
    with app.test_request_context(
        f"/query?q=extended||and|anything|contains|{text}&mode={lexicon}"
    ), mock.patch("karp5.context.auth.validate_user", return_value=(True, ["panacea"])):
        result = parser.parse(settings)

    expected = {
        "query": {
            "bool": {
                "must": [
                    {"term": {"lexiconName": lexicon}},
                    {"match": {"_all": {"query": text, "operator": "and"}}},
                ]
            }
        }
    }

    assert_es_search(result, expected)


def test_search_empty_input():
    exps = None
    filters = None
    fields = None

    result = parser.search(exps, filters, fields)

    expected = {}

    assert_es_search(result, expected)
    assert result == expected


def test_search_autocomplete():
    pass


@pytest.mark.parametrize(
    "exps,filters", [([{"match": {"field": "value"}}], [{"term": {"status": "ok"}}]),]
)
@pytest.mark.parametrize("constant_score", (True, False))
def test_search_case1(exps, filters, constant_score):
    fields = None
    isfilter = False
    highlight = False
    usefilter = True

    result = parser.search(exps, filters, fields, isfilter, highlight, usefilter, constant_score)
    if constant_score:
        expected = {"query": {"bool": {"filter": exps + filters}}}
        expected = {"query": {"constant_score": {"filter": {"bool": {"must": exps + filters}}}}}
    else:
        expected = {"query": {"bool": {"filter": filters, "must": exps}}}
        expected = {"bool": {"must": exps + filters}}

    assert_es_search(result, expected)
    assert result == expected


@pytest.mark.parametrize("exps", [([{"match": {"field": "value"}}]),])
@pytest.mark.parametrize("filters", [([{"term": {"status": "ok"}}]),])
@pytest.mark.parametrize("usefilter", [True, False])
@pytest.mark.parametrize("constant_score", [True, False])
def test_search_case2_list(exps, filters, usefilter, constant_score):
    fields = None
    isfilter = True
    highlight = False

    result = parser.search(exps, filters, fields, isfilter, highlight, usefilter, constant_score)

    expected = {"query": {"bool": {"filter": exps + filters}}}
    expected = {"filter": {"bool": {"must": exps + filters}}}
    assert_es_search(result, expected)

    assert result == expected


def test_search_case3_list():
    exps = [{"match": {"field": "value"}}]
    filters = [{"term": {"status": "ok"}}]
    fields = None
    isfilter = False
    highlight = False
    usefilter = False
    constant_score = True

    result = parser.search(exps, filters, fields, isfilter, highlight, usefilter, constant_score)

    expected = {"query": {"bool": {"filter": exps + filters}}}
    expected = {
        "bool": {
            "must": [{"match": {"field": "value"}}],
            "filter": {"bool": {"must": [{"term": {"status": "ok"}}]}},
        }
    }
    assert_es_search(result, expected)
    assert result == expected


def test_search_case3_list_no_constant_score():
    exps = [{"match": {"field": "value"}}]
    filters = [{"term": {"status": "ok"}}]
    fields = None
    isfilter = False
    highlight = False
    usefilter = False
    constant_score = False

    result = parser.search(exps, filters, fields, isfilter, highlight, usefilter, constant_score)

    expected = {"query": {"bool": {"filter": filters, "must": exps}}}
    expected = {
        "bool": {
            "must": [{"match": {"field": "value"}}],
            "filter": {"bool": {"must": [{"term": {"status": "ok"}}]}},
        }
    }

    assert_es_search(result, expected)
    assert result == expected


@pytest.mark.parametrize(
    "exps,filters", [([{"match": {"field": "value"}}], None), (None, [{"term": {"status": "ok"}}]),]
)
@pytest.mark.parametrize("constant_score", [True, False])
def test_search_case4(exps, filters, constant_score):
    fields = None
    isfilter = False
    highlight = False
    usefilter = False

    result = parser.search(exps, filters, fields, isfilter, highlight, usefilter, constant_score)
    if exps:
        expected = {"query": {"bool": {"filter" if constant_score else "must": exps}}}
        expected = {"query": {"bool": {"must": exps}}}
    else:
        expected = {}

    assert_es_search(result, expected)
    assert result == expected




from karp5.server.translator.parser import freetext

from karp5.tests.util import assert_es_search


def test_freetext_minimum():
    text = None
    mode = "karp"
    result = freetext(text, mode)

    expected = {
        "query": {
            "bool": {
                "should": [
                    {"match": {"_all": {"operator": "and", "query": text}}},
                    {"match": {"lemma_german": {"boost": 200, "query": text}}},
                    {"match": {"english.lemma_english": {"boost": 100, "query": text}}},
                ]
            }
        }
    }
    assert_es_search(result, expected)

    assert result == expected


def test_freetext_with_extra():
    text = None
    mode = "karp"
    extra = {"term": {"extra": "extra"}}
    result = freetext(text, mode, extra=extra)

    expected = {
        "query": {
            "bool": {
                "must": [
                    {
                        "bool": {
                            "should": [
                                {"match": {"_all": {"operator": "and", "query": text}}},
                                {"match": {"lemma_german": {"boost": 200, "query": text}}},
                                {"match": {"english.lemma_english": {"boost": 100, "query": text}}},
                            ]
                        }
                    },
                    extra,
                ]
            }
        }
    }
    assert_es_search(result, expected)

    assert result == expected


def test_freetext_with_filters():
    text = None
    mode = "karp"
    filters = [{"term": {"extra": "extra"}}]
    result = freetext(text, mode, filters=filters)

    expected = {
        "query": {
            "bool": {
                "should": [
                    {"match": {"_all": {"operator": "and", "query": text}}},
                    {"match": {"lemma_german": {"boost": 200, "query": text}}},
                    {"match": {"english.lemma_english": {"boost": 100, "query": text}}},
                ],
                "filter": filters[0],
            }
        }
    }

    assert_es_search(result, expected)
    assert result == expected




import logging
import os
import tempfile

from karp5 import Config


KARP5_DBPASS = os.environ.get("KARP5_DBPASS")

_tempfile = tempfile.NamedTemporaryFile(suffix=".db")


class TestConfig(Config):
    TESTING = True
    LOG_LEVEL = logging.DEBUG
    if KARP5_DBPASS is None:
        # Use sqlite if KARP5_DBPASS is not set
        # DATABASE_BASEURL = "sqlite://"
        DATABASE_BASEURL = f"sqlite:///{_tempfile.name}"
    ELASTICSEARCH_URL = os.environ.get("KARP5_ELASTICSEARCH_TEST_URL") or "localhost:9201"
    OVERRIDE_ELASTICSEARCH_URL = True



"""[summary]"""
import json
import logging
import os
import re
import copy
import sys
from typing import Dict, List, Optional, Union, Callable

import attr
from elasticsearch import Elasticsearch

# import karp5.server.helper.configpaths as C
from karp5 import errors

# from karp5.server.translator
from karp5.util.debug import print_err
from karp5.config.config import Config
from .errors import KarpConfigException


_logger = logging.getLogger("karp5")


def set_defaults(data: Dict) -> None:
    """[summary]

    Arguments:
        data {[type]} -- [description]
    """
    defaults = data.get("default", None)
    if not defaults:
        return
    print(f"data = {data}")
    for data_key, data_val in data.items():
        if data_key != "default":
            for def_key, def_val in defaults.items():
                if def_key not in data_val:
                    data_val[def_key] = def_val


# configdir = os.path.join(get_instance_path(), 'config')


def complement_dict(adict, bdict):
    " Adds values from bdict into adict unless the key is already there "
    for key, val in bdict.items():
        if key not in adict:
            adict[key] = val


def merge_dict(adict, bdict):
    " Merges bdict into adict by taking the union of the vaule lists "
    for key, val in bdict.items():
        if key in adict:
            adict[key] = list(set(adict[key]) | set(val))
        else:
            adict[key] = val


def load_from_file(path, default=None):
    """[summary]

    Arguments:
        path {[type]} -- [description]

    Keyword Arguments:
        default {[type]} -- [description] (default: {None})

    Returns:
        [type] -- [description]
    """
    try:
        with open(path) as fp:
            return json.load(fp)
    except Exception as e:
        print_err("Error when loading '{}':".format(path))
        print_err(e)

    if default is None:
        default = {}
    return default


# @attr.s(auto_attribs=True)
class ConfigManager(object):
    """[summary]

    Arguments:
        object {[type]} -- [description]

    Raises:
        KarpConfigException: [description]
        errors.KarpGeneralError: [description]

    Returns:
        [type] -- [description]
    """

    # instance_path: str = os.path.abspath(__file__)
    # version: str = ""
    # modes: Dict[str, Dict] = attr.Factory(dict)
    # config: Dict[str, Dict] = attr.Factory(dict)
    # lexicons: Dict[str, Dict] = attr.Factory(dict)
    # fields: Dict[str, Dict] = attr.Factory(dict)
    # defaultfields: Dict[str, Dict] = attr.Factory(dict)
    # _extra_src: Dict = attr.Factory(dict)
    # app_config: Config = attr.Factory(Config)
    # configdir: str = attr.ib(init=False)

    def __init__(self, instance_path):
        self.instance_path = instance_path
        self.version: str = ""
        self.config = {}
        self.lexicons = {}
        self.field = {}
        self.defaultfields = {}
        self.app_config: Config = Config()
        self.configdir = os.path.join(instance_path, "config")
        self._extra_src = {}
        self.load_config()

    # def __attrs_post_init__(self):
    #     self.configdir = os.path.join(self.instance_path, "config")
    #     self.load_config()

    def update_version(self, new_version: str):
        """Update the version string.

        Arguments:
            new_version {str} -- the new version to use
        """
        self.version = new_version

    def add_extra_src(self, mode, func):
        """[summary]

        Arguments:
            mode {[type]} -- [description]
            func {[type]} -- [description]
        """
        mode_fun_map = self._extra_src.get(mode, {})
        if func.__name__ in mode_fun_map:
            print_err(
                """WARNING!
                Function '{}' is already registered for mode '{}'. Overwritting...""".format(
                    func.__name__, mode
                )
            )

        mode_fun_map[func.__name__] = func
        self._extra_src[mode] = mode_fun_map

    def load_config(self):
        """[summary]
        """
        self.modes = load_from_file(os.path.join(self.configdir, "modes.json"))
        # print(f"self.modes = {self.modes}", file=sys.stderr)
        set_defaults(self.modes)

        self.lexicons = load_from_file(os.path.join(self.configdir, "lexiconconf.json"))
        set_defaults(self.lexicons)

        self.config = load_from_file(os.path.join(self.configdir, "config.json"))

        self.defaultfields = load_from_file(
            os.path.join(self.configdir, "mappings/fieldmappings_default.json")
        )

        # with open(os.path.join(self.configdir, 'fieldmappings.json')) as fp:
        #     self.fields = json.load(fp)
        self.fields = self.read_fieldmappings()

        # self.version = load_from_file(os.path.join(self.configdir, "karp5-version"))

    def read_fieldmappings(self):
        """[summary]

        Returns:
            [type] -- [description]
        """
        fields = {}
        for mode in self.modes.keys():
            fields[mode] = self.read_fieldmappings_mode(mode)
        return fields

    def read_fieldmappings_mode(self, mode):
        " Open a mode's config file and combine them "
        # " Default fields. Remember to add 'anything' to each index mapping "
        # defaultfields = _configmanager.defaultfields
        default_fields = copy.deepcopy(self.defaultfields)
        # step through the group members if the mode is an aliasmode
        # or use it's own config file if it's a indexmode
        all_fields = {}
        # print("create mode '{}'".format(mode))
        for index in self.modes[mode].get("groups", [mode]):
            try:
                # print("reading fieldmappings for index '{}'".format(index))
                with open(
                    os.path.join(self.configdir, "mappings/fieldmappings_{}.json".format(index))
                ) as fp:
                    fields = json.load(fp)
                merge_dict(all_fields, fields)
            except IOError:
                msg = "Couldn't find fieldmappings for mode '%s'"
                # print(msg)
                raise KarpConfigException(msg, index)

        complement_dict(all_fields, default_fields)
        return all_fields

    def override_elastic_url(self, elastic_url):
        """[summary]

        Arguments:
            elastic_url {[type]} -- [description]
        """
        for _, mode_settings in self.modes.items():
            mode_settings["elastic_url"] = elastic_url

    def filter_for_unauth_user(self, mode):
        return self.modes[mode].get("filter_for_unauth_user")

    def get_mode_sql(self, mode):
        """[summary]

        Arguments:
            mode {[type]} -- [description]

        Returns:
            [type] -- [description]
        """
        # dburl = 'mysql+pymysql://%s/%s?charset=utf8'
        dburl = self.app_config.DATABASE_BASEURL
        sql = self.searchconf(mode, "sql", failonerror=False)
        if sql:
            return dburl.format(sql)
            # return dburl % (C.config['DB']['DBPASS'], sql)
        else:
            return None

    def default_filename(self, lexicon):
        conf = self.lexicons.get(lexicon)
        if not conf:
            print(f"lexicons = {self.lexicons}", file=sys.stderr)
            print(f"configdir = {self.configdir}", file=sys.stderr)
            raise KarpConfigException("Can't find config for lexicon '%s'", lexicon)
        path = conf.get("path")
        if not path:
            _logger.warn(
                "Failed to load path for lexicon '%s'. Trying to read the 'default' path...",
                lexicon,
            )
            path = self.lexicons.get("default").get("path")
            if not path:
                msg = "Couldn't find 'path' for either '%s' or 'default'. Please check your config."
                _logger.error(msg, lexicon)
                raise KarpConfigException(msg, lexicon)

        fformat = conf.get("format", self.lexicons.get("format", "json"))
        return os.path.join(self.instance_path, path, f"{lexicon}.{fformat}")

    def searchconf(self, mode, field, failonerror=True) -> Optional[List[str]]:
        """[summary]

        Arguments:
            mode {[type]} -- [description]
            field {[type]} -- [description]

        Keyword Arguments:
            failonerror {bool} -- [description] (default: {True})

        Raises:
            errors.KarpGeneralError: [description]

        Returns:
            [type] -- [description]
        """
        # looks up field in modes.json, eg. "autocomplete"
        # returns the karp field name (eg. baseform.raw)
        try:
            _logger.debug("\n%s\n", self.modes[mode])
            return self.modes[mode][field]
        except Exception as e:
            if mode not in self.modes:
                msg = "Mode '%s' not found"
                _logger.error(msg, mode)
                _logger.exception(e)
                if failonerror:
                    raise errors.KarpGeneralError(msg % mode)
            else:
                msg = "Config field '%s' not found in mode '%s'"
                _logger.error(msg, field, mode)
                _logger.exception(e)
                if failonerror:
                    raise errors.KarpGeneralError(msg % (field, mode))
            return

    def extra_src(self, mode, funcname, default):
        """[summary]

        Arguments:
            mode {[type]} -- [description]
            funcname {[type]} -- [description]
            default {[type]} -- [description]

        Returns:
            [type] -- [description]
        """
        # Try in new way first
        mode_fun_map = self._extra_src.get(mode)
        if mode_fun_map:
            func = mode_fun_map.get(funcname)
            if func:
                _logger.debug("Found function '%s' for mode '%s'", funcname, mode)
                return func
            else:
                _logger.debug("Didn't find function '%s' for mode '%s'", funcname, mode)
        else:
            _logger.debug("Didn't find a function_map for mode '%s'", mode)

        # Then use old way
        import importlib

        # If importing fails, try with a different path.
        _logger.debug("look for %s in %s", funcname, mode)
        if mode not in self.modes:
            _logger.debug("Can't find mode '%s' in modes", mode)
        _logger.debug("sys.path = %s", sys.path)
        _logger.debug("file: %s", self.modes[mode]["src"])
        try:
            classmodule = importlib.import_module(self.modes[mode]["src"])
            _logger.debug("\n\ngo look in %s\n\n", classmodule)
            func = getattr(classmodule, funcname)
            return func
        except Exception as e:
            _logger.debug("Could not find %s in %s", funcname, self.modes[mode]["src"])
            _logger.debug(e)
            return default

    def elastic(self, mode="", lexicon=""):
        """[summary]

        Keyword Arguments:
            mode {str} -- [description] (default: {""})
            lexicon {str} -- [description] (default: {""})

        Returns:
            [type] -- [description]
        """
        return Elasticsearch(self.elasticnodes(mode=mode, lexicon=lexicon))

    def searchonefield(self, mode, field):
        """[summary]

        Arguments:
            mode {[type]} -- [description]
            field {[type]} -- [description]

        Returns:
            [type] -- [description]
        """
        # looks up field in modes.json, eg "autocomplete"
        # returns the first json path
        # TODO should change name to mode_one_field
        # TODO what to do if the field does not exist?
        # is probably handled elsewhere (searchconf or lookup_multiple)?
        return self.searchfield(mode, field)[0]

    def searchfield(self, mode, field):
        """[summary]

        Arguments:
            mode {[type]} -- [description]
            field {[type]} -- [description]

        Returns:
            [type] -- [description]
        """
        # looks up field in modes.json, eg "autocomplete"
        # returns the json path
        _logger.info("Looking for '%s' in '%s'", field, mode)
        fields = self.searchconf(mode, field)
        _logger.info("Found '%s' => '%s'", field, fields)
        return sum([self.lookup_multiple(f, mode) for f in fields], [])

    def all_searchfield(self, mode):
        # returns the json path of the field anything
        _logger.debug("%LOOK FOR ANYTHING\n")
        return self.lookup_multiple("anything", mode)

    def mode_fields(self, mode):
        return self.modes.get(mode, {})

    def formatquery(self, mode, field, op):
        format_query = self.extra_src(mode, "format_query", None)
        if format_query:
            return format_query(field, op)
        else:
            return op
        # return self.searchconf(mode, "format_query")(field, op)

    def elasticnodes(self, mode="", lexicon=""):
        if not mode:
            mode = self.get_lexicon_mode(lexicon)

        if self.app_config.OVERRIDE_ELASTICSEARCH_URL or "elastic_url" not in self.modes[mode]:
            return self.app_config.ELASTICSEARCH_URL
        else:
            return self.modes[mode]["elastic_url"]

    def get_lexicon_suggindex(self, lexicon):
        mode = self.get_lexicon_mode(lexicon)
        sugg_index = self.searchconf(mode, "suggestionalias")
        typ = self.searchconf(mode, "type")
        return sugg_index, typ

    def get_group_suggindex(self, mode):
        sugg_index = self.searchconf(mode, "suggestionalias")
        typ = self.searchconf(mode, "type")
        return sugg_index, typ

    def get_lexicon_index(self, lexicon):
        mode = self.get_lexicon_mode(lexicon)
        index = self.searchconf(mode, "indexalias")
        typ = self.searchconf(mode, "type")
        return index, typ

    def get_mode_type(self, mode):
        return self.searchconf(mode, "type")

    def get_mode_index(self, mode):
        index = self.searchconf(mode, "indexalias")
        typ = self.searchconf(mode, "type")
        return index, typ

    def get_lexicon_sql(self, lexicon):
        mode = self.get_lexicon_mode(lexicon)
        return self.get_mode_sql(mode)

    def get_lexiconlist(self, mode):
        lexiconlist = set()
        grouplist = [mode]
        modeconf = self.modes.get(mode, {})
        for group in modeconf.get("groups", []):
            grouplist.append(group)
        for lex, lexconf in self.lexicons.items():
            if lexconf.get("mode", "") in grouplist:
                lexiconlist.add(lex)

        return list(lexiconlist)

    def get_modes_that_include_mode(self, mode):
        modes = set()
        modes.add(mode)
        for name, info in self.modes.items():
            if mode in info.get("groups", []):
                modes.add(name)
        return list(modes)

    def get_lexicon_mode(self, lexicon: str) -> str:
        """Get mode that includes a lexicon.

        Arguments:
            lexicon -- the lexicon to query

        Returns:
            the mode that includes the lexicon
        """
        try:
            return self.lexicons[lexicon]["mode"]
        except Exception:
            # TODO what to return
            _logger.warning("Lexicon %s not in conf" % lexicon)
            return ""

    def get_mapping(self, index):
        filepath = "mappings/mappingconf_{}.json".format(index)
        try:
            with open(os.path.join(self.configdir, filepath)) as fp:
                return fp.read()
        except Exception:
            raise KarpConfigException(
                "Can't find mappingconf for index '%s' (looking in '%s')", index, self.configdir
            )

    def lookup(self, field, mode, own_fields=None):
        # standardmode = C.config['SETUP']['STANDARDMODE']
        # def lookup(self, field, mode=standardmode, own_fields={}):
        if own_fields is None:
            own_fields = {}
        return self.lookup_spec(field, mode, own_fields=own_fields)[0]

    def lookup_spec(self, field, mode, own_fields=None):
        # def lookup_spec(self, field, mode=standardmode, own_fields={}):
        if own_fields is None:
            own_fields = {}
        try:
            val = self.get_value(field, mode, own_fields=own_fields)
            if isinstance(val, dict):
                return ([val["search"]], (val["path"], val["typefield"], val["type"]))
            else:
                return (val[0],)
        except Exception as e:
            msg = "Field '%s' not found in mode '%s'" % (field, mode)
            _logger.error(msg + ": ")
            _logger.exception(e)
            raise errors.KarpGeneralError(msg)

    def lookup_multiple_spec(self, field, mode):
        # def lookup_multiple_spec(self, field, mode=standardmode):
        try:
            val = self.get_value(field, mode)
            if isinstance(val, dict):
                return ([val["search"]], (val["path"], val["typefield"], val["type"]))
            else:
                return (val, "")
        except Exception as e:
            msg = "Field '%s' not found in mode '%s'" % (field, mode)
            _logger.error(msg + ": ")
            _logger.exception(e)
            raise errors.KarpGeneralError(msg)

    def lookup_multiple(self, field, mode):
        # def lookup_multiple(self, field, mode=standardmode):
        _logger.info("Lookup '%s' in '%s'", field, mode)
        return self.lookup_multiple_spec(field, mode)[0]

    def get_value(self, field, mode, own_fields=None):
        if own_fields:
            use_fields = own_fields
        else:
            use_fields = self.fields
        mappings = use_fields.get(mode, {})
        group = re.search("(.*)((.bucket)|(.search)|(.sort))$", field)
        if field in mappings:
            return mappings[field]
        elif group is not None:
            return self.get_value(group.group(1), mode, own_fields=own_fields)
        else:
            msg = "Could not find field '%s' for mode '%s', '%s'"
            _logger.debug(msg, field, mode, mappings)
            raise KarpConfigException(msg, field, mode, mappings, debug_msg=msg)
